import json
from sparcur import schemas as sc


class RawJson:
    def __init__(self, path):
        self.path = path

    @property
    def data(self):
        with open(self.path, 'rt') as f:
            return json.load(f)


hasSchema = sc.HasSchema()
@hasSchema.mark
class RawJsonSubmission(RawJson):

    @hasSchema(sc.SubmissionSchema)
    def data(self):
        class RawSubmissionSchema(sc.JSONSchema):
            schema = sc.SubmissionSchema.schema['properties']['submission']

        rss = RawSubmissionSchema()
        blob = super().data
        try:
            rss.validate_strict(blob)
            # TODO this needs to be an error with an easy fix
            blob = {'submission': blob}
        except:
            pass

        return blob


hasSchema = sc.HasSchema()
@hasSchema.mark
class RawJsonDatasetDescription(RawJson):

    @hasSchema(sc.DatasetDescriptionSchema)
    def data(self):
        blob = super().data
        # TODO lift everything we can back to the ir
        class RawDatasetDescriptionSchema(sc.JSONSchema):
            schema = sc.DatasetDescriptionSchema.schema

        rds = RawDatasetDescriptionSchema()
        blob = super().data
        try:
            rds.validate_strict(blob)
        except:
            pass

        if not isinstance(blob['contributors'], list):
            # TODO this needs to be an error with an easy fix
            blob['contributors'] = [blob['contributors']]
            breakpoint()

        return blob


hasSchema = sc.HasSchema()
@hasSchema.mark
class RawJsonSubjects(RawJson):

    @hasSchema(sc.SubjectsSchema)
    def data(self):
        class RawSubjectsSchema(sc.JSONSchema):
            schema = sc.SubjectsSchema.schema['properties']['subjects']

        rss = RawSubjectsSchema()
        blob = super().data
        if isinstance(blob, list):
            # TODO this needs to be an error with an easy fix

            # try to do the right thing
            blob = {'subjects': blob}

        return blob



from collections import defaultdict
from functools import wraps
import idlib
from pyontutils.sheets import Sheet
from sparcur.core import adops, OntId, OntTerm
from sparcur.utils import log, logd
from urllib.parse import quote

# google sheets

# master
class Master(Sheet):
    name = 'sparc-master'


class Overview(Master):
    sheet_name = 'July18_2019'

    def dataset_ids(self):
        header, *rows = self.values
        index = header.index('identifier')
        return set(i for i in [row[index]for row in rows] if i)


class Progress(Master):
    sheet_name = 'Curation Progess (OT Only)'


class Grants(Master):
    sheet_name = 'Grant to Blackfynn to Protocols.io CLEAN'


class ISAN(Master):
    sheet_name = 'ISAN Demo July 2019'


class Participants(Master):
    sheet_name = 'Participan Notes'


class Protocols(Master):
    sheet_name = 'Protocol URL--> Blackfynn URL'


# affiliations
class Affiliations(Sheet):
    name = 'sparc-affiliations'
    sheet_name = 'Affiliations'
    index_columns = 'affiliation_string',

    def __call__(self, affiliation_string):
        if not isinstance(affiliation_string, str):
            logd.critical(str(affiliation_string))
            return self(affiliation_string[0] + 'ERROR ERROR')

        m = self.mapping

        if not isinstance(affiliation_string, str):
            log.critical('sigh')
            return None

        if affiliation_string in m:
            return m[affiliation_string]
        else:
            # FIXME super inefficient
            las = len(affiliation_string)
            for l, s in sorted([(len(k), k) for k in m.keys()], reverse=True):
                if l <= las and s in affiliation_string:
                    return m[s]

    @property
    def mapping(self):
        return {a:idlib.Ror(r) if r else None
                for a, r in zip(self.byCol.affiliation_string,
                                self.byCol.ror_id)}


# class sparc reports

class Reports(Sheet):
    """ spreadsheet that contains all spc reports as tabs """
    name = 'spc-reports'

    @classmethod
    def makeReportSheet(cls, *index_cols):
        """ Decorator to build a Sheet class for a cli.Report method

            NOTE reports that do not have index columns will update
            the entire sheet every time, so probabaly don't edit those
        """
        def _rowcellify(cell_value):
            if isinstance(cell_value, idlib.Stream):
                try:
                    return cell_value.asUri()
                except AttributeError as e:
                    breakpoint()
            elif isinstance(cell_value, str):
                if cell_value.lower() == 'x':
                    return True

            return cell_value

        def rowcellify(rows):
            return [[_rowcellify(cell_value) for cell_value in columns]
                    for columns in rows]

        def decorator(method):
            tags = method.__name__ == 'anno_tags'

            class ReportSheet(cls):
                __name__ = f'Report{method.__name__.capitalize()}'
                sheet_name = method.__name__
                index_columns = index_cols

            method._report_class = ReportSheet

            @wraps(method)
            def inner(self, *args, **kwargs):
                if 'ext' not in kwargs or kwargs['ext'] is None:
                    kwargs['ext'] = True
                else:
                    msg = 'Should not have extension and export to sheets.'
                    raise TypeError(msg)

                out = method(self, *args, **kwargs)

                if self.options.to_sheets:
                    if tags:
                        if args:
                            method._report_class.sheet_name = args[0]
                        else:
                            method._report_class.sheet_name = self.options.tag[0]

                    rows, title = out
                    # fetch=False since the remote sheet may not exist
                    # TODO improve the ergonimcs here
                    report = method._report_class(fetch=False, readonly=False)
                    report.metadata()  # idlib.Stream vs ? behavior not decided
                    rows_stringified = rowcellify(rows)
                    if report.sheetId() is None:
                        report.createRemoteSheet()
                        report.update(rows_stringified)
                    else:
                        report.fetch()
                        if report.index_columns:
                            report.upsert(*rows_stringified[1:])
                        else:
                            report.update(rows)

                    report.commit()

                return out

            return inner

        return decorator


class AnnoTags(Reports):
    name = 'anno-tags'


class WorkingExecVerb(AnnoTags):
    sheet_name = 'working-protc:executor-verb'
    index_columns = 'value',

    def condense(self):
        marked_as_done = '-'
        mapping = defaultdict(list)

        def make_key(row):
            return tuple(c.value for c in [row.tag(), row.value(), row.text(), row.exact()])

        create = []
        for mt_cell in self.row_object(0).map_to().column.cells[1:]:
            if (mt_cell.value and mt_cell.value != marked_as_done and
                mt_cell.value != mt_cell.row.value().value):
                try:
                    row, iv = self._row_from_index(value=mt_cell.value)
                    key = make_key(row)
                except AttributeError as e:  # value not in index
                    key = ('protc:executor-verb', mt_cell.value, '', '')
                    if key not in create:
                        create.append(key)
                        log.exception(e)

                mapping[key].append(mt_cell.row.value().value)  # cells don't move so we're ok

        mapping = dict(mapping)
        value_to_map_to = {value:k for k, values in mapping.items()
                           for value in values}

        #breakpoint()
        return value_to_map_to, create  # old -> new, original -> correct


# field alignment
class FieldAlignment(Sheet):
    name = 'sparc-field-alignment'


class Subjects(FieldAlignment):
    sheet_name = 'Subjects'


class Keywords(FieldAlignment):
    sheet_name = 'All Keywords'


class KeywordSets(FieldAlignment):
    sheet_name = 'Keyword sets'


class _Organs(FieldAlignment):
    # just having fetch_grid = True is enough to bolox everything
    sheet_name = 'Organs'
    index_columns = 'id',
    fetch_grid = True  # THIS IS THE CULPRET
    #fetch_grid = False  # THIS FLIES
    def modality(self, dataset_id): None
    def organ_term(self, dataset_id): None
    def award_manual(self, dataset_id): None
    def techniques(self, dataset_id): return []
    def protocol_uris(self, dataset_id): return []


class Organs(FieldAlignment):
    sheet_name = 'Organs'
    index_columns = 'id',
    fetch_grid = True
    #fetch_grid = False

    def _lookup(self, dataset_id, fail=False, raw=True):
        try:
            row = self.byCol.searchIndex('id', dataset_id, raw=raw)
            return row
        except KeyError as e:
            # TODO update the sheet automatically
            log.critical(f'New dataset! {dataset_id}')
            if fail:
                raise e

    def _dataset_row_index(self, dataset_id):
        # FIXME dict or cache this for performance
        row = self._lookup(dataset_id, fail=True)
        return self.values.index(row)

    def _update_dataset_metadata(self, id, name, award):
        try:
            row_index = self._dataset_row_index(id)
            row = self.row_object(row_index)

            cell_dsn = row.dataset_name()
            if cell_dsn.value != name:
                cell_dsn.value = name

            cell_award = row.award()
            if cell_award.value != award:
                cell_award.value = award

        except KeyError as e:
            row = ['', '', name, id, award]
            self._appendRow(row)

    def update_from_ir(self, ir):
        dataset_blobs = ir['datasets']
        self._wat = self.values[8]
        for blob in dataset_blobs:
            meta = blob['meta']
            self._update_dataset_metadata(
                id=blob['id'],
                name=adops.get(blob, ['meta', 'folder_name'], on_failure=''),
                award=adops.get(blob, ['meta', 'award_number'], on_failure=''),
            )

        #log.debug(self.uncommitted())
        self.commit()

    def modality(self, dataset_id):
        """ tuple of modalities """
        row = self._lookup(dataset_id)
        if row:
            try:
                m1 = self.byCol.header.index('modality1')
                m2 = self.byCol.header.index('modality2')
                return tuple(_ for _ in (row[m1], row[m2]) if _)
            except AttributeError as e:
                raise ValueError(f'issue in {self.name} {self.sheet_name}') from e

    def organ_term(self, dataset_id):
        row = self._lookup(dataset_id)
        organ_term = self.byCol.header.index('organ_term')
        if row:
            ot = row[organ_term] if row[organ_term] else None
            if ot:
                ts = tuple(OntId(t) for t in ot.split(' ') if t and t.lower() != 'na')
                return ts

    def award(self, dataset_id):
        row = self._lookup(dataset_id)
        award = self.byCol.header.index('award')
        award_manual = self.byCol.header.index('award_manual')
        if row:
            return row[award] if row[award] else (row[award_manual] if row[award_manual] else None)
        
    def award_machine(self, dataset_id):
        row = self._lookup(dataset_id)
        award = self.byCol.header.index('award')
        if row:
            return row[award] if row[award] else None

    def award_manual(self, dataset_id):
        row = self._lookup(dataset_id)
        award_manual = self.byCol.header.index('award_manual')
        if row:
            return row[award_manual] if row[award_manual] else None

    def techniques(self, dataset_id):
        row = self._lookup(dataset_id)

        def mkval(cell):
            hl = cell.hyperlink
            if hl is not None:
                oid = OntId(hl)
                if oid.prefix == 'TEMP':
                    logd.warning(f'{cell.value} -> {oid!r}')
                    #return OntTerm(curie=f'lex:{quote(cell.value)}')
                #else:

                return oid.asTerm()

            else:
                logd.warning(f'unhandled technique {cell.value}')
                return cell.value

        if row:
            row_index = self._dataset_row_index(dataset_id)
            return [mkval(self.cell_object(row_index, column_index))
                    for column_index, (field, value)
                    in enumerate(zip(self.byCol.header, row))
                    if field.startswith('technique') and
                    value and
                    value.lower() != 'na']
        else:
            return []

    def protocol_uris(self, dataset_id):
        row = self._lookup(dataset_id)

        def mkval(cell):
            hl = cell.hyperlink
            if hl is not None:
                return AutoId(hl)

            else:
                logd.warning(f'unhandled value {cell.value}')
                return cell.value

        if row:
            skip = '"none"', 'NA', 'no protocols', 'take protocol from other spreadsheet, '
            row_index = self._dataset_row_index(dataset_id)
            return [mkval(self.cell_object(row_index, column_index))
                    for column_index, (field, value)
                    in enumerate(zip(self.byCol.header, row))
                    if field.startswith('Protocol') and value and value not in skip]
        else:
            return []


def mkval_tech(cell):
    hl = cell.hyperlink
    if hl is not None:
        oid = OntId(hl)
        if oid.prefix == 'TEMP':
            logd.warning(f'{cell.value} -> {oid!r}')
            #return OntTerm(curie=f'lex:{quote(cell.value)}')
        #else:

        return oid.asTerm()

    else:
        logd.warning(f'unhandled technique {cell.value}')
        return cell.value


def mkval_pu(cell):
    hl = cell.hyperlink
    if hl is not None:
        return AutoId(hl)

    else:
        logd.warning(f'unhandled value {cell.value}')
        return cell.value



import os
import unittest
import pytest
from sparcur.paths import Path, BlackfynnCache as BFC
from sparcur.blackfynn_api import BFLocal
from sparcur.backends import BlackfynnRemote
from .common import test_organization, project_path_real


@pytest.mark.skipif('CI' in os.environ, reason='Requires access to data')
class TestBlackfynnRemote(unittest.TestCase):
    # FIXME skip in CI?
    def setUp(self):
        class BlackfynnCache(BFC):
            pass

        BlackfynnCache._bind_flavours()

        self.BlackfynnRemote = BlackfynnRemote._new(Path, BlackfynnCache)
        self.BlackfynnRemote.init(test_organization)
        if not project_path_real.exists():
            self.anchor = self.BlackfynnRemote.dropAnchor(project_path_real.parent)
        else:
            self.anchor = project_path_real.cache
            self.BlackfynnRemote.anchorTo(self.anchor)

        self.project_path = self.anchor.local

    def test_org(self):
        self.project_path.meta
        self.project_path.cache.meta
        self.project_path.remote.meta

        dsl = list(self.project_path.children)
        dsr = list(self.project_path.remote.children)

    def test_data(self):
        #dat = list(next(next(self.project_path.remote.children).children).data)
        dat = list(next(self.project_path.remote.children).data)
        #list(dd.data) == list(dd.remote.data)

    def test_children(self):
        #b = next(next(self.project_path.remote.children).children)
        b = next(self.project_path.remote.children)
        b.name

    def test_parts_relative_to(self):
        root = self.BlackfynnRemote(self.BlackfynnRemote.root)
        assert root.id == self.BlackfynnRemote.root



""" string normalizers, strings that change their content to match a standard """

import numbers
from types import GeneratorType
from html.parser import HTMLParser
import idlib
from pysercomb.pyr.types import Quantity
from . import exceptions as exc
from .core import log, logd, HasErrors
from .core import OntId
from .utils import is_list_or_tuple


BLANK_VALUE = object()
NOT_APPLICABLE = object()


class _Unknown(str):
    def __new__(cls, value=None):
        return str.__new__(cls, 'UNKNOWN')


UNKNOWN = _Unknown()


class NormSimple(str):

    data = tuple()  # really dict

    def __new__(cls, value):
        return str.__new__(cls, cls.normalize(value))

    @classmethod
    def normalize(cls, value, preserve_case=False):
        v = value if preserve_case else value.lower()

        if v in cls.data:
            return cls.data[v]
        elif preserve_case:
            return value
        else:
            return v


class NormAward(NormSimple):
    data = {
        '1 OT2 OD23853': 'OT2OD023853',  # someone's university database stripped a leading zero
        #'grantOT2OD02387101S2': '',
    }
    @classmethod
    def normalize(cls, value):
        _ovalue = value
        value = super().normalize(value, preserve_case=True)
        if 'OT2' in value and 'OD' not in value:
            # one is missing the OD >_<
            log.warning(value)
            value = value.replace('-', '-OD')  # hack

        n = (value
             .strip()
             .replace('-', '-')  # can you spot the difference?
             .replace('(', '')
             .replace(')', '')
             .replace('-01S1', '')
             .replace('-01', '')
             .replace('-02S2', '')
             .replace('-02', '')
             .replace('SPARC', '')
             .replace('NIH-1', '')
             .replace('NIH-', '')
             .replace('-', '')
             .replace('NIH ', '')
             .replace(' ', ''))
        if n[0] in ('1', '3', '5'):
            n = n[1:]

        if n.endswith('S2'):
            n = n[:-2]

        if n.endswith('D23864'):  # FIXME another trailing zero
            log.critical(_ovalue)
            n = n.replace('D23864', 'D023864')

        if n != _ovalue:
            log.debug(f'\n{_ovalue}\n{n}')
        return n


class NormFileSuffix(str):
    data = {
        'jpeg':'jpg',
        'tif':'tiff',
    }

    def __new__(cls, value):
        return str.__new__(cls, cls.normalize(value))

    @classmethod
    def normalize(cls, value):
        v = value.lower()
        ext = v[1:]
        if ext in cls.data:
            return '.' + data[ext]
        else:
            return v


class NormSpecies(NormSimple):
    data = {
        'cat':'Felis catus',
        'rat':'Rattus norvegicus',
        'mouse':'Mus musculus',
    }

class NormSex(NormSimple):
    data = {
        'm':'male',
        'f':'female',
    }


class NormHeader(NormSimple):
    __armi = 'age_range_min'
    __arma = 'age_range_max'
    data = {
        'age_range_minimum': __armi,
        'age_range_maximum': __arma,
        'protocol_io_location': 'protocol_url_or_doi',
        'protocol_io_location_2': 'protocol_url_or_doi',
        'protocol_title_2': 'protocol_title',
    }


class NormContributorRole(str):
    values = ('ContactPerson',
              'DataCollector',
              'DataCurator',
              'DataManager',
              'Distributor',
              'Editor',
              'HostingInstitution',
              'PrincipalInvestigator',  # added for sparc map to ProjectLeader probably?
              'CoInvestigator',  # added for sparc, to distingusih ResponsibleInvestigator
              'Creator',  # this is a separate field in datacite so we will need lift on export
              'Producer',
              'ProjectLeader',
              'ProjectManager',
              'ProjectMember',
              'RegistrationAgency',
              'RegistrationAuthority',
              'RelatedPerson',
              'Researcher',
              'ResearchGroup',
              'RightsHolder',
              'Sponsor',
              'Supervisor',
              'WorkPackageLeader',
              'Other',)

    def __new__(cls, value):
        return str.__new__(cls, cls.normalize(value))

    @staticmethod
    def levenshteinDistance(s1, s2):
        if len(s1) > len(s2):
            s1, s2 = s2, s1

        distances = range(len(s1) + 1)
        for i2, c2 in enumerate(s2):
            distances_ = [i2+1]
            for i1, c1 in enumerate(s1):
                if c1 == c2:
                    distances_.append(distances[i1])
                else:
                    distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
            distances = distances_
        return distances[-1]

    @classmethod
    def normalize(cls, value):
        # a hilariously slow way to do this
        # also not really normalization ... more, best guess for what people were shooting for
        if value:
            best = sorted((cls.levenshteinDistance(value, v), v) for v in cls.values)[0]
            distance = best[0]
            normalized = best[1]
            cutoff = len(value) / 2
            if distance > cutoff:
                msg = (f'"{value}" could not be normalized, best was {normalized} '
                       f'with distance {distance} cutoff was {cutoff}')
                raise exc.CouldNotNormalizeError(msg)

            return normalized


# static value normalization for complex inputs


class ATag(HTMLParser):
    text = None
    href = None
    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            for attr, value in attrs:
                if attr == 'href':
                    self.href = value

    def handle_endtag(self, tag):
        #print("Encountered an end tag :", tag)
        pass

    def handle_data(self, data):
        self.text = data
        #print("Encountered some data  :", data)

    def asJson(self, input):
        self.feed(input)
        if self.text is not None:
            return {'href': self.href, 'text': self.text}
        else:
            return self.href


class NormValues(HasErrors):
    """ Base class with an open dir to avoid name collisions """

    embed_bad_key_message = False  # TODO probably don't want this, better to zap bad values in pipeline

    def __init__(self, obj_inst):
        super().__init__()
        self._obj_inst = obj_inst

    def _bind(self):
        self._record_type_key_header = self._obj_inst.record_type_key_header
        self._norm_to_orig_alt = self._obj_inst.norm_to_orig_alt
        self._norm_to_orig_header = self._obj_inst.norm_to_orig_header
        self._groups_alt = self._obj_inst.groups_alt
        self._path = self._obj_inst.path

    def _error_on_na(self, value, key=None):
        """ N/A -> raise for cases where it should just be removed """
        if isinstance(value, str):
            v = value.strip()
            if v in ('NA', 'n/a', 'N/A',):
                # TODO consider double checking these cases ?
                raise exc.NotApplicableError(key)

    def _deatag(self, value):
        if value.startswith('<a') and value.endswith('</a>'):
            at = ATag()
            j = at.asJson(value)
            return at.href, j

        return value, None

    @staticmethod
    def _query(value, prefix):
        for query_type in ('term', 'search'):
            terms = list(OntTerm.query(prefix=prefix, **{query_type:value}))
            if terms:
                #print('matching', terms[0], value)
                #print('extra terms for', value, terms[1:])
                return terms[0]
            else:
                continue

        else:
            log.warning(f'No ontology id found for {value}')
            return value

    def _normv(self, thing, key=None, rec=None, path=tuple()):
        cell_errors = []
        if isinstance(thing, dict):
            out = {}
            for i, (k, v) in enumerate(thing.items()):
                intermediate = k in self._norm_to_orig_header
                process_group = k in self._groups_alt
                if k == self._record_type_key_header:
                    gnv = lambda : v
                elif intermediate:
                    gnv = lambda : self._normv(v, key, rec, path)
                elif k in self._norm_to_orig_alt:
                    gnv = lambda : self._normv(v, k, i, path + (k,))
                elif process_group:
                    gnv = lambda : self._normv(v, k, i, path + (k,))
                else:
                    raise ValueError(f'what is going on here?! {k} {v}')

                try:
                    nv = gnv()
                    out[k] = nv
                except exc.NotApplicableError:
                    pass
                except exc.TabularCellError as e:
                    # strip the error and log it
                    out[k] = e.value
                    # since out is mutable here it will continue to update
                    if rec is None or process_group:  # aka rec is None
                        if rec is None:
                            kmsg = f'.{k}'
                        else:
                            kmsg = f'.{key}'

                        if self.addError(str(e),
                                         pipeline_stage=f'{self.__class__.__name__}{kmsg}',
                                         blame='submission'):
                            log.critical(e)
                        #if isinstance(out[k], dict):
                            #self.embedLastError(out[k])

                    else:
                        if not intermediate and self.embed_bad_key_message:
                            self.errorInKey(out, k)

                        location = rec, i
                        new_e = exc.TabularCellError(str(e), value=out, location=location)
                        cell_errors.append(new_e)

            # end for
            if cell_errors:
                if len(cell_errors) != 1:
                    e = exc.TabularCellError(str(cell_errors),
                                             value=out,
                                             location=[e.location for e in cell_errors])
                else:
                    e = cell_errors[0]

                raise e 

            return out

        elif is_list_or_tuple(thing):
            # normal json not the tabular conversion case
            # or arrays at the bottom
            out = []
            _errors = []
            for i, v in enumerate(thing):
                try:
                    o = self._normv(v, key, i, path + (list,))
                    out.append(o)
                except exc.TabularCellError as e:
                    out.append(None)  # will be replaced in the if errors block
                    _errors.append((i, e))

            errors = [(i, e) for i, e in enumerate(out)
                      if isinstance(e, exc.TabularCellError)]
            errors = errors + _errors

            if errors:
                cell_errors = [e for i, e in errors]
                for i, e in errors:
                    out[i] = e.value

                e = exc.TabularCellError(str(cell_errors),
                                         value=out,
                                         location=[e.location for e in cell_errors])
                raise e

            return out

        else:
            # TODO make use of path
            # FIXME I do NOT like this pattern :/

            if isinstance(thing, str):
                self._error_on_na(thing, key)  # TODO see if this makes sense

            if isinstance(key, str) and hasattr(self, key):

                out = getattr(self, key)(thing)

                if isinstance(out, GeneratorType):
                    out = tuple(out)
                    errors = [(i, e) for i, e in enumerate(out)
                              if isinstance(e, exc.TabularCellError)]

                    if errors:
                        cell_errors = [e for i, e in errors]
                        out = [_ for _ in out]
                        for i, e in errors:
                            out[i] = e.value

                        out = tuple(out)
                        e = exc.TabularCellError(str(cell_errors),
                                                 value=out,
                                                 location=[e.location for e in cell_errors])
                        # FIXME this appears to cause double wrapping in tuples again >_<
                        raise e

                    if (len(out) == 1 and (key in self._obj_inst._expect_single or
                                           isinstance(out[0], Quantity))
                        or path[-1] == list):
                        # lists of lists might encounter issues here, but we almost never
                        # encounter those cases with metadata, especially in the tabular conversion
                        # but keep an eye out in which case we can check the schema type
                        # FIXME bad test around Quantity
                        out = out[0]
                    elif not out:
                        if isinstance(thing, str):
                            thing = thing.strip() # sigh

                        if thing is not None and thing != '':
                            msg = f'Normalization {key} returned None for input "{thing}"'
                            if self.addError(msg,
                                             pipeline_stage=f'{self.__class__.__name__}.{key}',
                                             blame='pipeline',):
                                log.critical(msg)

                        out = None

            else:
                out = thing

            if isinstance(out, exc.TabularCellError):
                raise out

            return out

    @property
    def data(self):
        #nk = self._obj_inst._normalize_keys()
        data_in = self._obj_inst._clean()
        self._bind()
        data_out = self._normv(data_in)
        self.embedErrors(data_out)
        return data_out



class NormSubmissionFile(NormValues):

    def milestone_achieved(self, value):
        # TODO and trigger na
        return value

    def sparc_award_number(self, value):
        return NormAward(value)


class NormDatasetDescriptionFile(NormValues):

    def additional_links(self, value):
        if value.startswith('<a') and value.endswith('</a>'):
            #return ATag().asJson(value)  # TODO not ready
            at = ATag()
            j = at.asJson(value)
            #return at.href, j
            return at.href

        return value

    def link_description(self, value):
        self._error_on_na(value)
        return value

    def prior_batch_number(self, value):
        self._error_on_na(value)
        return value

    def number_of_subjects(self, value):
        try:
            return int(value)
        except ValueError as e:
            if self.addError(str(e),
                             pipeline_stage=f'{self.__class__.__name__}',
                             blame='submission',):
                logd.exception(e)

            return value  # and let the schema sort them out

    number_of_samples = number_of_subjects

    def funding(self, value):
        if 'OT' in value:
            return NormAward(value)

        seps = '|', ';', ','  # order in priority
        for sep in seps:
            if sep in value:
                out = tuple()
                # FIXME duplicate values ?!
                for funding in value.split(sep):
                    out += (funding,)

                return out

        return value.strip()

    def __contributors(self, value):
        if isinstance(value, list):
            for d in value:
                yield {self.rename_key(k):tos(nv)
                    for k, v in d.items()
                    for nv in self.normalize(k, v) if nv}
        else:
            return value

    def contributor_affiliation(self, value):
        value, _j = self._deatag(value)
        return value

    def contributor_orcid_id(self, value):
        # FIXME use schema
        self._error_on_na(value)
        value, _j = self._deatag(value)

        v = value.replace(' ', '').strip().rstrip()  # ah the rando tabs in a csv
        if not v:
            return
        if v.startswith('http:'):
            v = v.replace('http:', 'https:', 1)

        if not (v.startswith('ORCID:') or v.startswith('https:')):
            v = v.strip()
            if not len(v):
                return
            elif v == '0':  # FIXME ? someone using strange conventions ...
                return
            elif len(v) != 19:
                msg = f'orcid wrong length {value!r} {self._path.as_posix()!r}'
                if self.addError(idlib.Orcid._id_class.OrcidLengthError(msg)):
                    logd.error(msg)

                return

            v = 'ORCID:' + v

        else:
            if v.startswith('https:'):
                _, numeric = v.rsplit('/', 1)
            elif v.startswith('ORCID:'):
                _, numeric = v.rsplit(':', 1)

            if not len(numeric):
                return
            elif len(numeric) != 19:
                msg = f'orcid wrong length {value!r} {self._path.as_posix()!r}'
                if self.addError(idlib.Orcid._id_class.OrcidLengthError(msg)):
                    logd.error(msg)
                return

        try:
            #log.debug(f"{v} '{self.path}'")
            orcid = idlib.Orcid(v)
            if not orcid.identifier.checksumValid:  # FIXME should not handle this with ifs ...
                # FIXME json schema can't do this ...
                msg = f'orcid failed checksum {value!r} {self._path.as_posix()!r}'
                if self.addError(idlib.Orcid._id_class.OrcidChecksumError(msg)):
                    logd.error(msg)
                return

            yield orcid

        except (OntId.BadCurieError, idlib.Orcid._id_class.OrcidMalformedError) as e:
            msg = f'orcid malformed {value!r} {self._path.as_posix()!r}'
            if self.addError(idlib.Orcid._id_class.OrcidMalformedError(msg)):
                logd.error(msg)
            yield value

    def contributor_role(self, value):
        # FIXME normalizing here momentarily to squash annoying errors
        cell_error = ''
        def echeck(original):
            orig = original.strip()
            try:
                return NormContributorRole(orig)
            except exc.CouldNotNormalizeError as e:
                #self.addError(e,
                              #pipeline_stage=f'{self.__class__.__name__}.contributor_role',
                              #logfunc=logd.error,
                              #blame='submission',
                              #path=self._path)
                emsg = f'Bad value: "{orig}"'
                nonlocal cell_error
                if cell_error:
                    cell_error = cell_error + ', ' + emsg
                else:
                    cell_error = emsg

                return orig

        def elist(vl):
            rv = tuple(sorted([o for o in set(echeck(e) for e in vl
                                              # have to filter here ourselves
                                              # since clean didn't function ...
                                              # unless ... we add an expand lists step ...
                                              # HRM, interesting
                                              if e is not None and e is not '')
                               if o]))
            nonlocal cell_error
            if cell_error:
                rv = tuple(exc.TabularCellError(cell_error, value=v) for v in rv)

            return rv

        if is_list_or_tuple(value):
            yield from elist(value)

        else:
            seps = '|', ';', ','  # order in priority
            for sep in seps:
                if sep in value:
                    break

            lst = value.split(sep)
            yield from elist(lst)

    def is_contact_person(self, value):
        # no truthy values only True itself
        yield value is True or isinstance(value, str) and value.lower() == 'yes'

    def _protocol_url_or_doi(self, value):
        doi = False
        if 'doi' in value:
            doi = True
        elif value.startswith('10.'):
            value = 'doi:' + value
            doi = True

        if doi:
            value = idlib.Doi(value)
        else:
            value = idlib.Pio(value)

        return value

    def title_for_complete_data_set(self, value):
        self._error_on_na(value)
        return value

    def protocol_url_or_doi(self, value):
        self._error_on_na(value)
        value, _j = self._deatag(value)

        for val in value.split(','):
            v = val.strip()
            if v:
                try:
                    yield self._protocol_url_or_doi(v)
                except BaseException as e:
                    #yield f'ERROR VALUE: {value}'  # FIXME not sure if this is a good idea ...
                    # it is not ...
                    _ps = f'{self.__class__.__name__}.protocol_url_or_doi'
                    if self.addError(str(e), pipeline_stage=_ps):
                        logd.error(e)

                    _pp = self._path.as_posix()
                    if self.addError(_pp,
                                     pipeline_stage=_ps,
                                     blame='debug'):
                        logd.critical(_pp)
                    # TODO raise exc.BadDataError from e

    def originating_article_doi(self, value):
        self._error_on_na(value)
        #self._error_on_tbd(value)  # TODO?
        value, _j = self._deatag(value)

        for val in value.split(','):
            v = val.strip()
            if v:
                try:
                    yield idlib.Doi(v)
                except idlib.exceptions.MalformedIdentifierError as e:
                    logd.exception(e)
                #doi = idlib.Doi(v)
                #if doi.valid:
                    # TODO make sure they resolve as well
                    # probably worth implementing this as part of OntId
                    #yield doi

    def keywords(self, value):
        if ';' in value:
            # FIXME error for this
            values = [v.strip() for v in value.split(';') if v]
        elif ',' in value:
            # FIXME error for this
            values = [v.strip() for v in value.split(',') if v]
        else:
            values = value,

        yield from values


class NormSubjectsFile(NormValues):

    _protocol_url_or_doi = NormDatasetDescriptionFile._protocol_url_or_doi
    protocol_url_or_doi = NormDatasetDescriptionFile.protocol_url_or_doi

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(self.__class__, 'pyru'):
            from pysercomb.pyr import units as pyru
            self.__class__.pyru = pyru

    def subject_id(self, value):
        if not isinstance(value, str):
            msg = f'Bad type for subject_id: {type(value)}'
            if self.addError(msg,
                             pipeline_stage=self.__class__.__name__,
                             blame='submission',):
                logd.error(msg)

            return str(value)
        else:
            return value

    def software_url(self, value):
        value, _j = self._deatag(value)
        return value

    def software_rrid(self, value):
        yield from self._rrid(value)

    def species(self, value):
        nv = NormSpecies(value)
        #yield self._query(nv, 'NCBITaxon')
        return nv

    def strain(self, value):
        if value == 'DSH':
            value = 'domestic shorthair'

        return value
        #wat = self._query(value, 'BIRNLEX')  # FIXME
        #yield wat

    sex = NormSex
    def sex(self, value):
        nv = NormSex(value)
        #yield self._query(nv, 'PATO')
        return nv

    def gender(self, value):
        # FIXME gender -> sex for animals, requires two pass normalization ...
        yield from self.sex(value)

    def group(self, value):
        # trigger n/a
        return value

    def pool_id(self, value):
        # trigger n/a
        return value

    def handedness(self, value):
        # needed to tirgger n/a fixes I think
        # TODO
        return value

    def _param(self, value):
        self._error_on_na(value)

        if isinstance(value, numbers.Number):
            yield self.pyru.ur.Quantity(value)
            return

        try:
            pv = self.pyru.UnitsParser(value).asPython()
        except self.pyru.UnitsParser.ParseFailure as e:
            caller_name = e.__traceback__.tb_frame.f_back.f_code.co_name
            msg = f'Unexpected and unhandled value "{value}" for {caller_name}'
            if self.addError(msg, pipeline_stage=self.__class__.__name__, blame='pipeline'):
                log.error(msg)

            if value.strip().lower() in ('unknown', 'uknown'):
                yield UNKNOWN
            else:
                yield value

            return

        #if not pv[0] == 'param:parse-failure':
        if pv is not None:  # parser failure  # FIXME check on this ...
            yield pv  # this one needs to be a string since it is combined below
        else:
            # TODO warn
            yield value

    def _param_unit(self, value, unit):
        if value.strip().lower() in ('unknown', 'uknown'):
            yield UNKNOWN
        else:
            yield from self._param(value + unit)

    def age(self, value):
        if value in ('adult',):
            msg = (f'Bad value for age: {value}\ndid you want to put that in age_cagegory instead?\n'
                   f'"{self._path}"')
            logd.error(msg)
            self.addError(msg, pipeline_stage=self.__class__.__name__, blame='submission',
                          path=self._path)
            return value

        yield from self._param(value)

    def age_years(self, value):
        # FIXME the proper way to do this is to detect
        # the units and lower them to the data, and leave the aspect
        yield from self._param_unit(value, 'years')

    #def age_category(self, value):
        #yield self._query(value, 'UBERON')

    def experimental_log_file_name(self, value):
        return value

    def age_range_min(self, value):
        yield from self._param(value)

    def age_range_max(self, value):
        if value in ('Normal',):
            msg = (f'Bad value for age_range_max: {value}\n'
                   'did you want to put that in age_cagegory instead?\n'
                   f'"{self._path}"')
            if self.addError(msg, pipeline_stage=self.__class__.__name__, blame='submission',
                             path=self._path):
                logd.error(msg)
            return value

        yield from self._param(value)

    age_range_max_disease = age_range_max # FIXME pretty sure these are a bad merge?

    def mass(self, value):
        yield from self._param(value)

    body_mass = mass

    def weight(self, value):
        yield from self._param(value)

    def weight_kg(self, value):  # TODO populate this?
        yield from self._param_unit(value, 'kg')

    def height_inches(self, value):
        yield from self._param_unit(value, 'in')

    def rrid_for_strain(self, value):
        yield from self._rrid(value)

    def _rrid(self, value):
        # OF COURSE RRIDS ARE SPECIAL
        if value.strip().lower() in ('unknown', 'uknown'):
            yield UNKNOWN
        else:
            try:
                rrid = idlib.Rrid(value)
                yield rrid
            except idlib.exceptions.MalformedIdentifierError as e:
                msg = f'malformed RRID: {value}'
                if self.addError(msg,
                                 pipeline_stage=self.__class__.__name__,
                                 blame='submission',
                                 path=self._path):
                    log.error(msg)

                if not value.startswith('RRID:'):
                    yield from self.rrid_for_strain('RRID:' + value)

    #def protocol_io_location(self, value):  # FIXME need to normalize this with dataset_description
        #yield value

    def _process_dict(self, dict_):
        """ deal with multiple fields """
        out = {k:v for k, v in dict_.items() if k not in self.skip}
        for h_unit, h_value in zip(self.h_unit, self.h_value):
            if h_value not in dict_:  # we drop null cells so if one of these was null then we have to skip it here too
                continue

            dhv = dict_[h_value]
            if isinstance(dhv, str):
                try:
                    dhv = ast.literal_eval(dhv)
                except ValueError as e:
                    raise exc.UnhandledTypeError(f'{h_value} {dhv!r} was not parsed!') from e

            compose = dhv * self.pyru.ur.parse_units(dict_[h_unit])
            #_, v, rest = parameter_expression(compose)
            #out[h_value] = str(UnitsParser(compose).for_text)  # FIXME sparc repr
            #breakpoint()
            out[h_value] = compose #UnitsParser(compose).asPython()

        if 'gender' in out and 'species' in out:
            if out['species'] != OntTerm('NCBITaxon:9606'):
                out['sex'] = out.pop('gender')

        return out

    #def experiment_date(self):
        #pass

    def ___iter__(self):
        """ this is still used """
        if self._is_json:
            yield from (self._process_dict({k:nv for k, v in d.items()
                                           for nv in self.normalize(k, v) if nv})
                        for d in self._data_raw)

        else:
            yield from (self._process_dict({k:nv for k, v in zip(r._fields, r) if v
                                           and k not in self.skip_cols
                                           for nv in self.normalize(k, v) if nv})
                        for r in self.bc.rows)

    def triples_gen(self, prefix_func):
        """ NOTE the subject is LOCAL """


class NormSamplesFile(NormSubjectsFile):

    def specimen_anatomical_location(self, value):
        seps = '|', ';'
        for sep in seps:
            if sep in value:
                for v in value.split(sep):
                    v = v.strip()
                    if v:
                        yield v
                return

        else:
            yield value.strip()



from docopt import parse_defaults
from sparcur import exceptions as exc
from sparcur.cli import Report, Options, __doc__ as clidoc
from sparcur.paths import Path, BlackfynnCache
from sparcur.config import auth
from sparcur.server import make_app
from sparcur.backends import BlackfynnRemote
from sparcur.curation import Summary

project_path = Path.cwd()
if not (project_path / Path._cache_class._local_data_dir).exists():
    raise exc.NotInProjectError(f'{project_path}')

defaults = {o.name:o.value if o.argcount else None for o in parse_defaults(clidoc)}
args = {'server': True,
        '--raw': False,
        '--latest': True,
        '--sort-count-desc': True,
        '--project-path': project_path,
        '--tab-table': False,
        '<path>': [],
        '--verbose': False,

        '--export-path': auth.get_path('export-path'),
        '--partial': False,
        '--open': False,
        '--debug': False,
}

options = Options(args, defaults)
report = Report(options)

# set report paths that would normally be populated from Main
report.cwd = options.project_path
report.project_path = options.project_path
report.project_id = project_path.cache.id  # FIXME should not have to do this manually?
report.anchor = project_path.cache
report.summary = Summary(options.project_path)
report._timestamp = None  # FIXME
report._folder_timestamp = None  # FIXME

# set up bfapi
report.BlackfynnRemote = BlackfynnRemote._new(Path, BlackfynnCache)
report.BlackfynnRemote.init(report.project_id)

app, *_ = make_app(report, project_path)
app.debug = False

if __name__ == '__main__':
    app.run(host='localhost', port=defaults['--port'], threaded=True)



import logging
import hashlib
import pathlib
from functools import wraps
from itertools import chain
try:
    import magic  # from sys-apps/file consider python-magic ?
except ImportError:
    pass

import augpathlib as aug
from dateutil import parser
from augpathlib import PrimaryCache, EatCache, SqliteCache, SymlinkCache
from augpathlib import RepoPath, LocalPath
from sparcur import backends
from sparcur import exceptions as exc
from sparcur.utils import log


def cleanup(func):
    @wraps(func)
    def inner(self):
        meta = func(self)
        if meta is not None:
            if meta.id.startswith('EUf5'):
                if self.is_broken_symlink():
                    self.unlink()
                elif self.exists() and not self.is_symlink():
                    self.unlink()
                    self.touch()
                else:
                    pass

            else:
                return meta
            
    return inner


class BlackfynnCache(PrimaryCache, EatCache):
    xattr_prefix = 'bf'
    _backup_cache = SqliteCache
    _not_exists_cache = SymlinkCache

    cypher = hashlib.sha256  # set the remote hash cypher on the cache class

    uri_human = backends.BlackfynnRemote.uri_human
    uri_api = backends.BlackfynnRemote.uri_api

    @classmethod
    def decode_value(cls, field, value):
        if field in ('created', 'updated'):
            # if you get unicode decode error here it is because
            # struct packing of timestamp vs isoformat are fighting
            # in xattrs pathmeta helper
            return value.decode()  # path meta handles decoding for us

    @property
    def anchor(self):
        return self.organization

    @property
    def _id_parts(self):
        ntid = self.id
        if ntid is None:
            raise exc.NoCachedMetadataError(self)

        N, thing, id = ntid.split(':')
        return N, thing, id

    @property
    def _id_uuid(self):
        N, thing, id = self._id_parts
        return id  # TODO inverse function

    @property
    def _fs_safe_id(self):
        N, thing, id = self._id_parts
        return thing[0] + '-' + id

    @property
    def _trashed_path(self):
        sid = self._fs_safe_id
        suuid = self._id_uuid
        try:
            pid = self.parent._fs_safe_id
        except exc.NoCachedMetadataError as e:
            msg = f'Projects cannnot be trashed when trying to trash {self}'
            raise TypeError(msg) from e

        fid = f'{self.file_id}-' if self.file_id else ''
        # we use sid as the sparsification folder here to ensure
        # a more uniform distribution of deletions across folders
        # therefore if you need to find files deleted from a specific folder
        # use find -name '*collection-id-part*'
        return self.trash  / suuid[:2] / f'{sid}-{pid}-{fid}{self.name}'

    @property
    def _trashed_path_short(self):
        return self.trash / 'short' / self.name

    @property
    def trash(self):
        return self.local_data_dir / 'trash'

    @property
    def organization(self):
        """ organization represents a permissioning boundary
            for blackfynn, so above this we would have to know
            in advance the id and have api keys for it and the
            containing folder would have some non-blackfynn notes
            also it seems likely that the same data could appear in
            multiple different orgs, so that would mean linking locally
        
            NOTE: unlike RemotePath, CachePath should use the local file
            structure to search for the local anchor.
        """

        # FIXME in reality files can have more than one org ...
        if self.meta is not None:
            id = self.id
            if id is None:
                parent = self.parent
                if parent == self:  # we have hit a root
                    return None

                organization = parent.organization

                if organization is not None:
                    # TODO repair
                    pass
                else:
                    raise exc.NotInProjectError

            elif id.startswith('N:organization:'):
                return self

        if self.parent and self.parent != self:
            # we have a case of missing metadata here as well
            return self.parent.organization

    def is_organization(self):  # FIXME ?
        return self.id and self.id.startswith('N:organization:')

    def is_dataset(self):
        return self.id and self.id.startswith('N:dataset:')

    @property
    def dataset(self):
        if self.is_dataset():
            return self

        elif self.parent and self.parent != self:  # Path('.') issue
            log.debug(self.parent)
            return self.parent.dataset

    @property
    def dataset_id(self):
        dataset = self.dataset
        if dataset:
            return dataset.id

    @property
    def file_id(self):
        return self.meta.file_id

    @property
    def cache_key(self):
        return f'{self.id}-{self.file_id}'

    def _dataset_metadata(self, force_cache=False):
        """ get metadata about a dataset from the remote metadata store """
        # TODO figure out the right default for force_cache
        dataset = self.dataset
        if dataset == self:
            if not hasattr(self, '_c__dataset_metadata'):
                bdd = BlackfynnDatasetData(self)
                try:
                    blob = bdd.fromCache()
                except FileNotFoundError as e:
                    # FIXME TODO also check cached rmeta dates during pull
                    if force_cache:
                        raise e
                    else:
                        log.warning(e)
                        blob = bdd()

                self._c__dataset_metadata = blob

            return self._c__dataset_metadata

        else:
            return dataset._dataset_metadata()

    def _package_count(self):
        if self.is_dataset():
            return sum(self._dataset_metadata()['package_counts'].values())
        else:
            raise NotImplementedError('unused at the moment')

    def _sparse_materialize(self, *args, sparse_limit=None):
        """ use data from the remote mark or clear datasets as sparse """
        if sparse_limit is None:
            sparse_limit = auth.get('sparse-limit')  # yay for yaml having int type
            #sparse_limit = _sl if _sl is None else int_(sl)

        if self.is_dataset():
            package_count = self._package_count()
            sparse_remote = (False
                             if sparse_limit is None else
                             package_count >= sparse_limit)
            sparse_cache = self.is_sparse()
            if sparse_remote:
                if not sparse_cache:
                    self._mark_sparse()
            elif sparse_cache:  # strange case where number of packages decreases
                self._clear_sparse()

        else:
            msg = 'at the moment only datasets can be marked as sparse'
            raise NotImplementedError(msg)

    @property
    def data(self):
        """ get the 'cached' data which isn't really cached at the moment
            once we implement an index for local files then we can hit that
            first from here """
        # we don't keep two copies of the local data
        # unless we are doing a git-like thing
        if self.is_dir():
            raise TypeError('can\'t retrieve data for a directory')

        meta = self.meta
        if meta.file_id is None:
            raise NotImplementedError('can\'t fetch data without a file id')

        #cands = list(self.local_object_cache_dir.glob(self.cache_key))
        # FIXME this does not play well with old_id ...
        # can probably get away with just globing for the old_id in
        # most cases
        # TODO where to store the chain of prior versions? i.e. do
        # we just keep the xattrs in the object cache? how about file moves?
        # sigh git ...
        if self.local_object_cache_path.exists():
            gen = chain((f'from local cache {self.local_object_cache_path}',),
                        self.local_object_cache_path.data)
        else:
            gen = self._remote_class.get_file_by_id(meta.id, meta.file_id)

        try:
            self.data_headers = next(gen)
        except exc.NoRemoteFileWithThatIdError as e:
            log.error(f'{self} {e}')
            raise exc.CacheNotFoundError(f'{self}') from e  # have to raise so that we don't overwrite the file

        log.debug(self.data_headers)
        if self.local_object_cache_path.exists():
            yield from gen
        else:
            yield from self.local_object_cache_path._data_setter(gen)
            self.local_object_cache_path.cache_init(self.meta)  # FIXME self.meta be stale here?!

    def _meta_is_root(self, meta):
        return meta.id.startswith('N:organization:')

    def _bootstrap_recursive(self, only=tuple(), skip=tuple(), sparse=False):
        """ only on the first call to this function should sparse be a tuple """
        # bootstrap the rest if we told it to
        if self.id.startswith('N:organization:'):  # FIXME :/
            yield from self.remote.children_pull(self.children,
                                                 only=only,
                                                 skip=skip,
                                                 sparse=sparse,)

        else:
            if isinstance(sparse, list) and self.id.startswith('N:dataset:'):
                sparse = self.id in sparse

            yield from super()._bootstrap_recursive(sparse=sparse)

    _sparse_stems = (
        'manifest', 'dataset_description', 'submission', 'subjects', 'samples'
    )

    def _sparse_include(self):
        sl = self.stem.lower()
        return bool([an for an in self._sparse_stems if an in sl])

    def populateJsonMetadata(self, blob):
        """ populate a json blob with file system metadata"""
        # TODO schema for this
        if blob is None:
            blob = {}

        blob.update(self._jsonMetadata())

    def _jsonMetadata(self, do_expensive_operations=False):
        """ path level json metadata """
        # FIXME this is going to be very slow due to the local.cache call overhead
        # it will be much better implement this from Path directly using xattrs()
        drp = self.local.relative_to(self.dataset.local)
        meta = self.meta  # TODO see what we need from this
        N, package = self.id.split(':', 1)
        uri_api = self.uri_api  # FIXME vs self.uri_api_package ?
        uri_human = self.uri_human,
        blob = {
            'type': 'path',
            'uri_api': uri_api,  # -> @id in the @context
            'uri_human': uri_human,
            'package_id': package,  # FIXME N:package:asdf is nasty for jsonld but ...
            'dataset_relative_path': drp,
        }

        mimetype = self.mimetype
        if mimetype:
            blob['mimetype'] = mimetype

        if do_expensive_operations:
            blob['magic_mimetype'] = asdf

        return blob


BlackfynnCache._bind_flavours()


class Path(aug.XopenPath, aug.RepoPath, aug.LocalPath):  # NOTE this is a hack to keep everything consistent
    """ An augmented path for all the various local
        needs of the curation process. """

    _cache_class = BlackfynnCache

    @property
    def cache_id(self):
        """ fast way to get cache.id terrifyingly this has roughly
        an order of magnitude less overhead than self.cache.id """

        try:
            return (self.getxattr('bf.id').decode()
                    if self.is_dir() or self.is_file() else
                    self.readlink().parts[1])
        except OSError as e:
            raise exc.NoCachedMetadataError(self) from e

    def populateJsonMetadata(self, blob):
        """ populate a json blob with file system metadata"""
        # FIXME TODO implement this without resolving .cache
        return self.cache.populateJsonMetadata(blob)

    def pull(self, *args,
             time_now=None,
             debug=False,
             n_jobs=12,
             cache_anchor=None,
             log_name=None,
             log_level='INFO',
             # pass the paralle in at run time if needed
             Parallel=None,
             delayed=None,
             _in_parallel=False,):
        # TODO usage errors

        if _in_parallel:
            _log = logging.getLogger(log_name)
            _log.setLevel(log_level)
            rc = self._remote_class
            if not hasattr(rc, '_cache_anchor'):
                rc.anchorTo(cache_anchor)

        else:
            _log = log

        cache = self.cache

        if cache.is_organization():
            if debug or Parallel is None:
                for child in self.children:
                    child.pull()
            else:
                Parallel(n_jobs=n_jobs)(
                    delayed(child.pull)(_in_parallel=True,
                                        time_now=time_now,
                                        cache_anchor=cache.anchor,
                                        log_name=_log.name,
                                        log_level=log_level,)
                    for child in self.children)

        elif cache.is_dataset():
            self._pull_dataset(time_now)  # XXX actual pull happens in here

        else:
            raise NotImplementedError(self)

    def _pull_dataset(self, time_now):
        cache = self.cache
        if cache.is_organization():
            raise TypeError('can\' use this method on organizations')
        elif not cache.is_dataset():
            return self.dataset._pull_dataset()

        children = list(self.children)
        if not children:
            return list(self.cache.rchildren)

        # instantiate a temporary staging area for pull
        ldd = cache.local_data_dir
        contain_upstream = ldd / 'temp-upstream'
        contain_upstream.mkdir(exist_ok=True)

        # create the new directory with the xattrs needed to pull
        working = self
        upstream = contain_upstream / working.name
        upstream.mkdir()
        upstream.setxattrs(working.xattrs())  # handy way to copy xattrs

        # FIXME TODO error handling

        # pull and prep for comparison of relative paths
        upstream_c_children = list(upstream.cache.rchildren)  # XXX actual pull happens here
        upstream_children = [c.local.relative_to(upstream)
                            for c in upstream_c_children]
        working_children = [c.relative_to(working)
                            for c in working.rchildren]

        # FIXME TODO comparison/diff and support for non-curation workflows

        # replace the working tree with the upstream ca
        upstream.swap_carefree(working)

        upstream_now_old = upstream

        # neither of these cleanup approaches is remotely desireable
        suf = f'-{time_now.START_TIMESTAMP_LOCAL_SAFE}'
        upstream_now_old.rename(upstream_now_old.parent /
                                (upstream_now_old.name + suf))  # FIXME

    def updated_cache_transitive(self):
        """ fast get the date for the most recently updated cached path """
        if self.cache.is_organization():
            gen = (rc for c in self.children for rc in c.rchildren)
        elif self.cache.is_dataset():
            gen = self.rchildren
        else:
            gen = chain((self,), self.rchildren)

        simple_meta = [aug.PathMeta(updated=c.getxattr('bf.updated').decode())
                       if not c.is_broken_symlink() else aug.PathMeta.from_symlink(c)
                       for c in gen]
        if simple_meta:
            return max(m.updated for m in simple_meta)

    def upload(self, replace=True, local_backup=False):
        # FIXME we really need a staging area ...
        remote, old_remote = (
            self
            ._cache_class
            ._remote_class
            ._stream_from_local(self,
                                replace=True,
                                local_backup=False))

        if True:  # local_backup = True:  # force True until we can get remote checksums
            # FIXME there must be a better way to do this ...
            object_path = self._cache_class._anchor.local_objects_dir / remote.cache_key
            # TODO possibly use _data_setter to calculate the hash at the same time?
            object_path.copy_from(self)
            object_path.cache_init(remote.meta)

        if self.cache is None:
            # FIXME didn't we already figure out the right way to do this?
            cache = self.cache_init(remote.meta)
            self._cache = cache
            cache._remote = remote
            remote._cache = cache
        else:
            # remote has no cache and self already has a cache in this case
            # so we can't just assign directly so we have to say which
            # cache we want to update, and make sure that names match
            # the cache should already have been updated EXCEPT for the
            # fact that the remote fileid is not known, but we are inside
            # the code that actually makes that update ...
            # and the absense of a staging area / history makes the cache.meta
            # the only place we can do this besides the object store, especially
            # if we lack the ability to retrieve checksums for single files (sigh)

            # we already have the latest data (ignoring concurency)
            remote.update_cache(cache=self.cache, fetch=False) # FIXME fetch=False => different diff rule

        return remote


Path._bind_flavours()


class StashCache(BlackfynnCache):
    def remote(self):
        return None

    def _meta_setter(self, value):
        if not self.meta:
            super()._meta_setter(value)
        else:
            raise TypeError('you dont want to set a stashed cache')


class StashPath(Path):
    _cache_class = StashCache
    @property
    def remote(self):
        return None


# assign defaults

BlackfynnCache._local_class = Path
backends.BlackfynnRemote._new(Path, BlackfynnCache)
backends.BlackfynnRemote.cache_key = BlackfynnCache.cache_key
backends.BlackfynnRemote._sparse_stems = BlackfynnCache._sparse_stems
backends.BlackfynnRemote._sparse_include = BlackfynnCache._sparse_include

# end imports (woo circular deps)
from sparcur.datasources import BlackfynnDatasetData



import ast
import copy
import json
import types
import inspect
from functools import wraps
import jsonschema

# FIXME these imports should not be here types rules should be set in another way
from pathlib import PurePath
import idlib
import rdflib
import requests
import ontquery as oq
from pysercomb.pyr.types import ProtcurExpression
from sparcur import exceptions as exc
from sparcur.utils import logd
from sparcur.core import JEncode, JApplyRecursive, JPointer, OntCuries
from pyontutils.utils import asStr
from pyontutils.namespaces import TEMP, TEMPRAW, sparc, unit, PREFIXES as uPREFIXES, ilxtr


CONTEXT_ROOT = object()  # used to mark where nested context_value statements should be deposited
prefix_endswith = ['/', '#', '_', '-', ':', '=',
                   '/fma', '/sao', 'PTHR', '/chebi#2', '/chebi#3',]

base_context = {
    '@version': 1.1,
    'id': '@id',
    'dataset': {'@id': 'https://api.blackfynn.io/datasets/N:dataset:', '@prefix': True},
    'package': {'@id': 'https://api.blackfynn.io/packages/N:package:', '@prefix': True},
    # @base and _bfc are added at runtime and are dataset:{dataset-id-suffix}
    'meta': '@graph',
    #'meta': 'TEMP:hasSubGraph',  #'_bfc:#meta-graph',  # FIXME I think the fragment is the right thing to do here ...
    'subjects': 'TEMP:hasSubGraph',  #'_bfc:#subjects-graph',  # FIXME I think the fragment is the right thing to do here ...
    'samples': 'TEMP:hasSubGraph',  #'_bfc:#samples-graph',
    'contributors': '_bfc:#contributors-graph',
    #'N:dataset': {'@id': 'https://api.blackfynn.io/datasets/'},  # darn it multiprefix issues
    **{p: {'@id': n, '@prefix': True}
       if [c for c in prefix_endswith if n.endswith(c)]
       else n for p, n in {**uPREFIXES, **OntCuries._dict}.items()
       if p != ''},  # FIXME massive data duplication
    'unit': {'@id': 'TEMP:hasUnit',
             '@type': '@id',
             '@context': {'@base': str(unit)}},
    'value': {'@id': 'rdf:value'},  # FIXME maybe include this in a dedicated quantity @context?
    'type': {'@id': 'ilxtr:jsonRecordType',  # FIXME ...
             '@type': '@id',
             '@context': {'@base': str(ilxtr['sparcur/types/'])},},
    'quantity': 'sparc:Measurement',  # for some reason this isn't working, and is captured by @base
    #'identifier': 'owl:Class',
    'system': {'@id': 'ilxtr:identifierType',  # FIXME see what predicate we used for this
               '@type': '@id',
               '@context': {'@base': str(ilxtr['idlib/systems/'])},},
    'OntTerm': 'owl:Class',
    #'uri_api': {'@id': '@id', '@type': '@id'},  # klobbered below, now fixed
}


def json_version(version):
    """ version should be a branch or a release tag e.g. schema-0.0.1 """
    return {
        '$schema': 'http://json-schema.org/draft-06/schema#',
        '$id': ('https://raw.githubusercontent.com/SciCrunch/sparc-curation/'
                f'{version}/resources/defs.schema.json'),
        'title': 'Common definitions for the SDS json schemas, including mappings to json-ld @context',
        'description': 'JSON-LD mappings are under the `context_value` key by convention',
        '@context': {
            'sparc': str(sparc),
            'TEMP': str(TEMP),
            'TEMPRAW': str(TEMPRAW),
        },
        'definitions': {
            'patterns': {
                'metadata_filename': metadata_filename_pattern,
                'simple_url': simple_url_pattern ,
                #'doi': idlib.Doi._id_class.canonical_regex,
                #'orcid': idlib.Orcid._id_class.canonical_regex,
                #'ror': idlib.Ror._id_class.canonical_regex,
            'iso8601_date': iso8601datepattern,
                'iso8601_datetime': iso8601pattern,
                'whitespace_lead_trail': pattern_whitespace_lead_trail,
            },
            'NoLTWhitespace': NoLTWhitespaceSchema.schema,
            'EmbeddedIdentifier': EmbeddedIdentifierSchema.schema,

        'meta': MetaOutExportSchema.schema,
            'prov': ProvSchema.schema,
            'errors': ErrorSchema.schema,
            'contributor': ContributorOutExportSchema.schema,
            'subject': SubjectExportSchema.schema,
            'sample': SampleExportSchema.schema,
            'dataset_description_file': DatasetDescriptionSchema.schema,
            'submission_file': SubmissionSchema.schema,
            'subjects_file': SubjectsSchema.schema,
            'samples_file': SamplesFileSchema.schema,
            'dataset': {
                'meta': {'$ref': '#/definitions/meta'},
                'prov': {'$ref': '#/definitions/prov'},
                'errors': {'$ref': '#/definitions/errors'},
                'contributor': {'$ref': '#/definitions/contributor'},
                'subject': {'$ref': '#/definitions/subject'},
                'sample': {'$ref': '#/definitions/sample'},
                'dataaset_description_file': {'$ref': '#/definitions/dataaset_description_file'},
                'submission_file': {'$ref': '#/definitions/submission_file'},
                'subjects_file': {'$ref': '#/definitions/subjects_file'},
                'samples_file': {'$ref': '#/definitions/samples_file'},
                'errors': {'$ref': '#/definitions/errors'},
            }
        }
    }


def _defunc(obj, *args, path=None, **kwargs):  # FIXME add tests
    if isinstance(obj, types.FunctionType):
        return asStr(ast.parse(inspect.getsource(obj).strip().strip(',')))
    else:
        return obj


def update_include_paths(obj, *args, path=None, key='jsonld_include', moves=tuple(), collect=tuple()):
    """ collect json pointers that need to be updated """
    if isinstance(obj, dict) and key in obj:
        path = list(JPointer.pathFromSchema(path))
        for frm, to in moves:  # sigh python
            lf = len(frm)
            if path[:lf] == frm:
                path = to + path[lf:]
                break

        for key, value in obj[key].items():
            path.append(key)
            collect.append((path, lambda _: value))  # note that collect is really a list
            # note also that lambda _: value is because pipeline adds currently calls into lifters

    return obj  # have to return this otherwise somehow everything is turned to None?


def idtype(id, base=None):
    out = {'@id': id, '@type': '@id'}
    if base:
        out['@context'] = {'@base': str(base)}

    return out


def strcont(id):
    return {'type': 'string', 'context_value': id}


def inttype(id):
    return {'@id': id, '@type': 'xsd:integer'}


def not_array(schema, in_all=False):
    """ hacked way to determine that the blob in question cannot be an array """
    return ('type' in schema and schema['type'] != 'array' or
            'properties' in schema and 'items' not in schema or
            # FIXME hack hard to maintain
            'allOf' in schema and all(not_array(s, True) for s in schema['allOf']) or
            'oneOf' in schema and all(not_array(s, True) for s in schema['oneOf']) or
            'anyOf' in schema and all(not_array(s, True) for s in schema['anyOf']) or
            'not' in schema and (not not_array(schema['not'], in_all) or in_all)
    )


class hproperty:
    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        if self.fget is None:
            raise AttributeError("unreadable attribute")

        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")

        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")

        self.fdel(obj)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


class HasSchema:
    """ decorator for classes with methods whose output can be validated by jsonschema """
    def __init__(self, input_schema_class=None, fail=True, normalize=False):
        self.input_schema_class = input_schema_class
        self.fail = fail
        self.normalize = normalize
        self.schema = None  # deprecated output schema ...

    def mark(self, cls):
        """ note that this runs AFTER all the methods """

        if self.input_schema_class is not None:
            # FIXME probably better to do this as types
            # and check that schemas match at class time
            cls._pipeline_start = cls.pipeline_start

            @hproperty
            def pipeline_start(self, isc=self.input_schema_class, fail=self.fail):
                if pipeline_start.schema == isc:
                    pipeline_start.schema = isc()

                schema = pipeline_start.schema
                data = self._pipeline_start
                ok, norm_or_error, data = schema.validate(data)
                if not ok and fail:
                    raise norm_or_error

                return data

            pipeline_start.schema = self.input_schema_class
            cls.pipeline_start = pipeline_start

        return cls

        # pretty sure this functionality is no longer used
        if self.schema is not None:
            cls._output = cls.output
            @property
            def output(_self):
                return self.schema.validate(cls._output)

            cls.output = output

        return cls

    def f(self, schema_class, fail=False):
        return self(schema_class, fail=fail, _property=False)

    def __call__(self, schema_class, fail=False, _property=True):
        # TODO switch for normalized output if value passes?
        def decorator(function):
            pipeline_stage_name = function.__qualname__
            @wraps(function)
            def schema_wrapped_property(_self, *args, **kwargs):
                if schema_wrapped_property.schema == schema_class:
                    schema_wrapped_property.schema = schema_class()

                schema = schema_wrapped_property.schema
                data = function(_self, *args, **kwargs)
                ok, norm_or_error, data = schema.validate(data)
                if not ok:
                    if fail:
                        logd.error('schema validation has failed and fail=True')
                        raise norm_or_error

                    try:
                        if 'errors' not in data:
                            data['errors'] = []
                    except BaseException as e:
                        raise exc.SparCurError(
                            f'Error from {_self.__class__.__name__}.'
                            f'{function.__name__}') from e
                        
                    data['errors'] += norm_or_error.json(pipeline_stage_name)
                    # TODO make sure the step is noted even if the schema is the same

                if self.normalize:
                    return norm_or_error

                return data

            if _property:
                schema_wrapped_property = hproperty(schema_wrapped_property)

            schema_wrapped_property.schema = schema_class
            return schema_wrapped_property

        return decorator


class ConvertingChecker(jsonschema.FormatChecker):
    _enc = JEncode()

    def check(self, instance, format):
        converted = _enc.default(instance)
        return super().check(converted, format)


class JSONSchema(object):

    schema = {}

    validator_class = jsonschema.Draft6Validator

    string_types = [  # add more types here
        str,
        PurePath,
        idlib.Stream,  # FIXME str/dict
        rdflib.URIRef,
        rdflib.Literal,
        oq.OntId,
        oq.OntTerm,  # FIXME str/dict
        ProtcurExpression,
    ]

    def __init__(self):
        format_checker = jsonschema.FormatChecker()
        #format_checker = ConvertingChecker()
        types = dict(array=(list, tuple),
                     string=tuple(self.string_types))
        self.validator = self.validator_class(self.schema,
                                              format_checker=format_checker,
                                              types=types)

    @classmethod
    def _add_meta(cls, blob):
        """ In place modification of a schema blob to add schema level metadata """
        # TODO schema metadata, I don't know enough to do this correctly yet
        #schema_version = cls.validator_class.ID_OF(cls.validator_class.META_SCHEMA)
        #schema['$schema'] = schema_version

        #schema_id = '#/temp/' + cls.__name__ + '.json'  # FIXME version ...
        #schema['$id'] = schema_id

    @classmethod
    def export(cls, base_path):
        """ write schema to file with """
        def pop_errors(dict_):
            if 'errors' in dict_:
                dict_.pop('errors')
            for v in dict_.values():
                if isinstance(v, dict):
                    pop_errors(v)

        schema = copy.deepcopy(cls.schema)
        pop_errors(schema)
        cls._add_meta(schema)

        def debug(obj, *args, path=None, **kwargs):
            if isinstance(obj, types.FunctionType):
                raise TypeError(path + [obj])
            else:
                return obj

        schema = JApplyRecursive(_defunc, schema)
        JApplyRecursive(debug, schema)

        with open((base_path / cls.__name__).with_suffix('.json'), 'wt') as f:
            json.dump(schema, f, sort_keys=True, indent=2)

    def validate_strict(self, data):
        # Take a copy to ensure we don't modify what we were passed.
        #appstruct = copy.deepcopy(data)

        appstruct = json.loads(json.dumps(data, cls=JEncode))  # FIXME figure out converters ...

        errors = list(self.validator.iter_errors(appstruct))
        if errors:
            raise exc.ValidationError(errors)

        return appstruct

    def validate(self, data):
        """ capture errors """
        try:
            ok = self.validate_strict(data)  # validate {} to get better error messages
            return True, ok, data  # FIXME better format

        except exc.ValidationError as e:
            return False, e, data  # FIXME better format

    @classmethod
    def context(cls):
        """ return a json-ld context by extracting annotations from nested schemas
            those annotations are currently placed under the `context_value' key """
        # FIXME last one wins right now
        #context_runtimes = {}
        context = {}
        marker = 'context_value'
        #marker_runtime = 'context_runtime'
        def inner(schema, key=None):
            if isinstance(schema, dict):
                if marker in schema:
                    if key is None:
                        raise ValueError('you probably don\'t want to do this ...')

                    context_value = schema[marker]
                    context[key] = context_value

                #if marker_runtime in schema:
                    #if key is None:
                        #raise ValueError('you probably don\'t want to do this ...')

                    #context_runtimes[key] = schema[marker_runtime]

                for k, v in schema.items():
                    if k == marker:
                        continue
                    inner(v, k)

            elif isinstance(schema, list):
                for ss in schema:
                    inner(ss, key)

            else:
                pass

        inner(cls.schema)

        #if context:  # if there is no context do not add version
            #context.update(base_context)

        #if context_runtimes:
            #def context_runtime(base):
                #return {k: v(base) for k, v in context_runtimes.items()}

        return context


class JsonLdHelperSchema(JSONSchema):
    """ documentation of the additions to the json schema draft validator schema
        that are used for transformation to jsonld """
    schema = {
        'properties': {
            'jsonld_includes': {'type': 'object'},
            'context_value': {'oneOf': [
                # re https://github.com/json-ld/json-ld.org/issues/612
                # https://github.com/json-ld/json-ld.org/blob/master/schemas/jsonld-schema.json
                {'type': 'string'},  # TODO are there any restrictions? string -> @base/string ?
                {'type': 'object'},  # TODO this should conform to the jsonld schema for @context
            ]}
        }
    }


class RuntimeSchema(JSONSchema):
    """ finally a zero (runtime) cost first time only init
    with reasonable debugability and readability """

    schema = None

    def __new__(cls):
        return super().__new__(cls)

    _renew = __new__

    def __new__(cls, *args, **kwargs):
        cls.setup(*args, **kwargs)
        cls.__new__ = cls._renew
        return super().__new__(cls)

    def __init__(self):
        super().__init__()

    _reinit = __init__

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.__class__.__init__ = self.__class__._reinit

    @classmethod
    def setup(cls):
        raise NotImplementedError('subclassit')


class RemoteSchema(RuntimeSchema):
    schema = ''
    # FIXME TODO these need to be cached locally
    @classmethod
    def setup(cls):
        cls.schema = requests.get(cls.schema).json()


class ApiNATOMYSchema(RemoteSchema):
    schema = ('https://raw.githubusercontent.com/open-physiology/'
              'open-physiology-viewer/master/src/model/graphScheme.json')


class ToExport(RuntimeSchema):
    """ version should match the github release tag i.e. schema-0.0.1 """

    @classmethod
    def setup(cls, version=None):
        if version is None:
            version = 'master'
        #cls.schema = JApplyRecursive(_defunc, json_version(version))
        cls.schema = json_version(version)


metadata_filename_pattern = r'^.+\/[a-z_\/]+\.(xlsx|csv|tsv|json)$'

simple_url_pattern = r'^(https?):\/\/([^\s\/]+)\/([^\s]*)'

# NOTE don't use builtin date-time format due to , vs . issue
iso8601pattern = '^[0-9]{4}-[0-1][0-9]-[0-3][0-9]T[0-2][0-9]:[0-6][0-9]:[0-6][0-9](,[0-9]{6})*(Z|[-\+[0-2][0-9]:[0-6][0-9]])'
iso8601datepattern = '^[0-9]{4}-[0-1][0-9]-[0-3][0-9]'

# https://www.crossref.org/blog/dois-and-matching-regular-expressions/
doi_pattern = idlib.Doi._id_class.canonical_regex

orcid_pattern = idlib.Orcid._id_class.canonical_regex

ror_pattern = idlib.Ror._id_class.canonical_regex

pattern_whitespace_lead_trail = '(^[\s]+[^\s].*|.*[^\s][\s]+$)'


class NoLTWhitespaceSchema(JSONSchema):
    schema = {'allOf': [{'type': 'string'},
                        {'not': {'type': 'string',
                                 'pattern': pattern_whitespace_lead_trail,
                        }},]}


string_noltws = NoLTWhitespaceSchema.schema


class ErrorSchema(JSONSchema):
    schema = {'type':'array',
              'minItems': 1,
              'items': {'type': 'object'},}



def make_id_schema(cls, pattern=None, format=None):
    id_schema = {'type': 'string'}
    if format is not None:
        id_schema['format'] = format

    if pattern is not None:
        id_schema['pattern'] = pattern
    elif hasattr(cls, '_id_class') and hasattr(cls._id_class, 'canonical_regex'):
        id_schema['pattern'] = cls._id_class.canonical_regex

    class _IdSchema(JSONSchema):
        __name__ = cls.__name__ + 'Schema'
        schema = {'properties': {'system': {'type': 'string',
                                            'enum': [cls.__name__],},
                                 'id': id_schema},}

    return _IdSchema


OntTermSchema = make_id_schema(oq.OntTerm, format='iri')
DoiSchema = make_id_schema(idlib.Doi)
RorSchema = make_id_schema(idlib.Ror)
OrcidSchema = make_id_schema(idlib.Orcid)
PioSchema = make_id_schema(idlib.Pio)
RridSchema = make_id_schema(idlib.Rrid)


class EmbeddedIdentifierSchema(JSONSchema):
    schema = {'type': 'object',
              'required': ['type', 'id', 'label'],
              'properties': {
                  'id': {'type': 'string'},
                  'type': {'type': 'string',
                           'enum': ['identifier']},
                  'system': {'type': 'string'},
                  'label': {'type': 'string'},
                  'synonyms': {'type': 'array',
                               'minItems': 1},
                  # NOTE: description is the normalized form name ofr
                  # ontology term definitions
                  'description': {'type': 'string'},},}

    @classmethod
    def _to_pattern(cls, obj, *args, path=None, **kwargs):  # FIXME add tests
        if (isinstance(obj, dict) and
            'allOf' in obj and
            (len(obj) == 1 or len(obj) == 2 and 'context_value' in obj) and  # FIXME extremely brittle
            len(obj['allOf']) == 2 and
            obj['allOf'][0] == cls.schema):
            return obj['allOf'][1]['properties']['id']

        else:
            return obj

    @classmethod
    def _allOf(cls, schema, context_value=None):
        ao = {'allOf': [cls.schema, schema.schema,]}
        if context_value:
            ao['context_value'] = context_value

        return ao

    @classmethod
    def allOf(cls, schema_):
        class _SCM(JSONSchema):
            schema = cls._allOf(schema_)  # LOL PYTHON

        return _SCM


EIS = EmbeddedIdentifierSchema
EISs = EmbeddedIdentifierSchema.schema


class ProvSchema(JSONSchema):
    schema = {'type': 'object',
              'properties': {'timestamp_export_start': {'type': 'string',
                                                        'pattern': iso8601pattern,},
                             #'timestamp_export_end': {'type': 'string'},  # this isn't really possible
                             'export_system_identifier': {'type': 'string'},
                             'export_system_hostname': {'type': 'string'},
                             'export_project_path': {'type': 'string'},
                             'sparcur_version': {'type': 'string'},
                             'sparcur_commit': {'type': 'string'},
                             'ontology_mappings': {
                                 'type': 'array',
                                 'minItems': 1,
                                 'items': {
                                     'type': 'object',
                                     'properties': {
                                         'curie': {'type': 'string'},
                                         'label': {'type': 'string'},
                                         'matches': {
                                             'type': 'array',
                                             'minItems': 1,
                                             'items': {
                                                 'type': 'object',
                                                 'properties': {
                                                     'path': {'type': 'array'},
                                                     'input': {'type': 'string'},
                                                 },},},},},},},}


class MbfTracingSchema(JSONSchema):
    schema = {
        'type': 'object',
        'required': ['subject', 'atlas', 'images', 'contours'],
        'properties': {
            'subject': {'type': 'object',
                        'required': ['id'],
                        'properties': {'id': {'type': 'string'},
                                       'species': {'type': 'string'},
                                       'sex': {'type': 'string'},
                                       'age': {'type': 'string'},
                                       },},
            'atlas': {'type': 'object',
                      'properties': {'organ': {'type': 'string'},
                                     'atlas_label': {'type': 'string'},
                                     'atlas_rootid': {'type': 'string'},
                                     },},
            'images': {'type': 'array',
                       'items': {'type': 'object',
                                 'properties':
                                 {'path_mbf': {
                                     'type': 'array',
                                     'minItems': 1,
                                     'items': {'type': 'string'}},
                                  'channels': {
                                      'type': 'array',
                                      'items': {
                                          'type': 'object',
                                          'properties': {
                                              'id': {'type': 'string'},
                                              'source': {'type': 'string'},
                                          },},},},},},
            'contours': {'type': 'array',
                         'items': {'type': 'object',
                                   'properties': {
                                       'name': string_noltws,
                                       'guid': {'type': 'string'},
                                       'id_ontology': {'type': 'string'},},
                                   },},},}


class NeurolucidaSchema(JSONSchema):
    schema = copy.deepcopy(MbfTracingSchema.schema)
    schema['required'] = ['images', 'contours']


class DatasetStructureSchema(JSONSchema):
    __schema = {'type': 'object',
                'required': ['submission_file', 'dataset_description_file'],
                'properties': {
                    'submission_file': {'type': 'string',
                                        'pattern': metadata_filename_pattern},
                    'dataset_description_file': {'type': 'string',
                                                 'pattern': metadata_filename_pattern},
                    'subjects_file': {'type': 'string',
                                      'pattern': metadata_filename_pattern},
                    'samples_file': {'type': 'string',
                                     'pattern': metadata_filename_pattern},
                    'manifest_file': {'type': 'array',
                                      'minItems': 1,
                                      'items': {
                                          'type': 'string',
                                          'pattern': metadata_filename_pattern},
                                      },
                    #'path_structure': {'type': 'array',
                    # 'items': {'type': 'string', 'pattern': ''}}
                    'dirs': {'type': 'integer',
                             'minimum': 0},
                    'files': {'type': 'integer',
                              'minimum': 0},
                    'size': {'type': 'integer',
                             'minimum': 0},
                    'errors': ErrorSchema.schema,
                }}
    schema = {'allOf': [__schema,
                        {'anyOf': [
                            {'required': ['subjects_file']},
                            {'required': ['samples_file']}]}]}


class ContributorExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Person', 'owl:NamedIndividual']},
        'properties': {
            'id': {'type': 'string',  # FIXME TODO
                   },
            'affiliation': EIS._allOf(RorSchema,
                                      context_value=idtype('TEMP:hasAffiliation')),
            'contributor_name': {'type': 'string'},
            'first_name': {'type': 'string',
                           'context_value': 'sparc:firstName',},
            'middle_name': strcont('TEMP:middleName'),
            'last_name': {'type': 'string',
                          'context_value': 'sparc:lastName',},
            'contributor_orcid_id': EIS._allOf(OrcidSchema, context_value=idtype('sparc:hasORCIDId')),
            'blackfynn_user_id': strcont(idtype('TEMP:hasBlackfynnUserId')),
            'contributor_affiliation': {'anyOf': [EIS._allOf(RorSchema),
                                                  {'type': 'string'}],
                                        'context_value': idtype('TEMP:hasAffiliation')
                                        },
            'contributor_role': {
                'type':'array',
                'context_value': idtype('TEMP:hasRole', base=str(TEMP)),
                'minItems': 1,
                'items': {
                    'type': 'string',
                    'enum': ['ContactPerson',
                             'Creator',  # allowed here, moved later
                             'DataCollector',
                             'DataCurator',
                             'DataManager',
                             'Distributor',
                             'Editor',
                             'HostingInstitution',
                             'PrincipalInvestigator',  # added for sparc map to ProjectLeader probably?
                             'CoInvestigator', # added for sparc, to distingusih ResponsibleInvestigator
                             'Producer',
                             'ProjectLeader',
                             'ProjectManager',
                             'ProjectMember',
                             'RegistrationAgency',
                             'RegistrationAuthority',
                             'RelatedPerson',
                             'Researcher',
                             'ResearchGroup',
                             'RightsHolder',
                             'Sponsor',
                             'Supervisor',
                             'WorkPackageLeader',
                             'Other',]}
            },
            'is_contact_person': {'type': 'boolean'},
            'errors': ErrorSchema.schema,
        }}


class ContributorSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(ContributorExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class ContributorsExportSchema(JSONSchema):
    schema = {'type': 'array',
              #'context_runtime': lambda base: {'@id': f'{base}contributors',},
              'contains': {
                  'type': 'object',
                  'required': ['is_contact_person'],
                  'properties': {
                      'is_contact_person': {'type': 'boolean', 'enum': [True]},
                  },
              },
              'items': ContributorExportSchema.schema,
            }

class ContributorsSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(ContributorsExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class ContributorOutExportSchema(JSONSchema):
    schema = copy.deepcopy(ContributorExportSchema.schema)
    schema['properties']['contributor_role']['items']['enum'].remove('Creator')


class ContributorsOutExportSchema(JSONSchema):
    schema = copy.deepcopy(ContributorsExportSchema.schema)
    schema['items'] = ContributorOutExportSchema.schema


class CreatorExportSchema(JSONSchema):
    schema = copy.deepcopy(ContributorExportSchema.schema)
    schema.pop('jsonld_include')
    schema['properties'].pop('contributor_role')
    schema['properties'].pop('is_contact_person')


class CreatorSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(CreatorExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)



class CreatorsExportSchema(JSONSchema):
    schema = {'type': 'array',
              'minItems': 1,
              'items': CreatorExportSchema.schema}


class CreatorsSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(CreatorsExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


_protocol_url_or_doi_schema = {'anyOf':[EIS._allOf(DoiSchema),
                                        EIS._allOf(PioSchema),
                                        {'type': 'string',
                                         'pattern': simple_url_pattern,}]}


class DatasetDescriptionExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'additionalProperties': False,
        'required': ['schema_version',  # missing should fail for this one ...
                     'name',
                     'description',
                     'funding',
                     'protocol_url_or_doi',
                     #'completeness_of_data_set'  # TODO versioned schema validation
                     'contributors',
                     'number_of_subjects',
                     'number_of_samples',],
        # TODO dependency to have title for complete if completeness is is not complete?
        'properties': {
            'errors': ErrorSchema.schema,
            'schema_version': {'type': 'string'},
            'name': string_noltws,
            'description': {'type': 'string'},
            'keywords': {'type': 'array', 'items': {'type': 'string'}},
            'acknowledgements': {'type': 'string'},
            'funding': {'type': 'array',
                        'minItems': 1,
                        'items': {'type': 'string'}},
            'completeness_of_data_set': {'type': 'string'},
            'prior_batch_number': {'type': 'string'},
            'title_for_complete_data_set': {'type': 'string'},
            'originating_article_doi': {'type': 'array',
                                        'items': EIS._allOf(DoiSchema),
                                        #'context_value': idtype('TEMP:hasDoi'),  # FIXME convert here?
                                        },
            'number_of_subjects': {'type': 'integer'},
            'number_of_samples': {'type': 'integer'},
            'parent_dataset_id': {'type': 'string'},  # blackfynn id
            'protocol_url_or_doi': {'type': 'array',
                                    'minItems': 1,
                                    'items': _protocol_url_or_doi_schema,
                                    #'context_value': idtype('TEMP:hasProtocol'),  # FIXME convert here?
                                    },
            'links': {
                'type': 'array',
                'minItems': 1,
                'items': {
                    'type': 'object',
                    'properties': {
                        'additional_links': {'type': 'string'},
                        'link_description': {'type': 'string'},  # FIXME only if there is an additional link? some using for protocol desc
                    }
                }
            },
            'examples': {
                'type': 'array',
                'minItems': 1,   # this pattern makes things self describing
                'items': {
                    'type': 'object',
                    'properties': {
                        'example_image_filename': {'type': 'string'},
                        'example_image_locator': {'type': 'string'},
                        'example_image_description': {'type': 'string'},
                    }
                }
            },
            'contributors': ContributorsExportSchema.schema,
        }
    }


class DatasetDescriptionSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(DatasetDescriptionExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class SubmissionSchema(JSONSchema):
    context = lambda : ({}, None)
    schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'errors': ErrorSchema.schema,
            'submission': {
                'type': 'object',
                'required': ['sparc_award_number',
                             'milestone_achieved',
                             'milestone_completion_date'],
                # TODO allOf?
                'properties': {'sparc_award_number': {'type': 'string'},
                               'milestone_achieved': {'type': 'string'},
                               'milestone_completion_date': {'type': 'string',
                                                             'context_value': 'TEMP:milestoneCompletionDate',
                                                             'pattern': iso8601datepattern,},}}}}


class UnitSchema(JSONSchema):
    schema = {'oneOf': [{'type': ['number', 'string']},
                        {'type': 'object',
                         # TODO enum on the allowed values here
                         'properties': {'type': {'type': 'string'},
                                        'value': {'type': 'number'},
                                        'unit': {'type': 'string'}}}]}


_software_schema = {'type': 'array',
                    'minItems': 1,
                    'items': {
                        'type': 'object',
                        'properties': {
                            'software_version': {'type': 'string'},
                            'software_vendor': {'type': 'string'},
                            'software_url': {'type': 'string'},
                            'software_rrid': EIS._allOf(RridSchema),
                        },},}


subsam_common_properties = {  # FIXME these should not be in the samples sheet, or if they are they refer to the sample
    'species': strcont('sparc:animalSubjectIsOfSpecies'),
    'strain': strcont('sparc:animalSubjectIsOfStrain'),
    'rrid_for_strain': EIS._allOf(RridSchema, context_value=idtype('TEMP:specimenRRID')),  # XXX CHANGE from sparc:specimenHasIdentifier
    'genus': strcont('sparc:animalSubjectIsOfGenus'),
    'sex': {'type': 'string',
            'context_value': 'TEMP:hasBiologicalSex',  # FIXME idtype
            },  # sex as a variable ?

    'body_mass':{'context_value': 'sparc:animalSubjectHasWeight',
                 **UnitSchema.schema},
    'body_weight': {'context_value': 'sparc:animalSubjectHasWeight',
                    **UnitSchema.schema},
    'age': {'context_value': 'TEMP:hasAge',
            **UnitSchema.schema},
    'age_years': {'context_value': 'TEMP:hasAge',
                  **UnitSchema.schema},
    'age_category': {'type': 'string',
                     'context_value': idtype('TEMP:hasAgeCategory'),
                     },  # TODO uberon
    'age_range_min': {'context_value': 'TEMP:hasAgeMin',
                      **UnitSchema.schema},
    'age_range_max': {'context_value': 'TEMP:hasAgeMax',
                      **UnitSchema.schema},
    'handedness': strcont('TEMP:hasHandedness'),

    # ambiguous
    # genotype is ambiguous when dealing with tissue from chimeric subjects or experimentally manipulated cell populations
    'genotype': strcont('TEMP:hasGenotype'),
    'mass': {'context_value': 'sparc:animalSubjectHasWeight',
             # XXX NOTE subjects will require their own context
             # since mass can clearly have other meanings elsewhere
             **UnitSchema.schema},
    'weight': {'context_value': 'sparc:animalSubjectHasWeight',
               **UnitSchema.schema},
    'experimental_group': strcont('TEMP:hasAssignedGroup'),
    'group': strcont('TEMP:hasAssignedGroup'),
    'organism_rrid': strcont('TEMP:organismRRID'),

    # fields that require additional information
    'experiment_date': strcont('TEMP:protocolExecutionDate'),  # FIXME needs to reference a protocol
    'injection_date': strcont('TEMP:protocolExecutionDate'),  # FIXME needs to reference a protocol
    'date_of_euthanasia': strcont('TEMP:protocolExecutionDate'),
    'date_of_injection': strcont('TEMP:protocolExecutionDate'),

    # fields that are actually about files
    'sha1': strcont('TEMP:hasDigitalArtifactThatIsAboutItWithHash'),
    'filename': strcont('TEMP:hasDigitalArtifactThatIsAboutIt'),
    'upload_filename': strcont('TEMP:hasDigitalArtifactThatIsAboutIt'),
    'dataset_id': strcont('TEMP:providerDatasetIdentifier'),  # FIXME need a global convention for this

    # fields that are actually about performances
    'experiment_number': strcont('TEMP:localPerformanceNumber'),  # FIXME TODO
    'session': strcont('TEMP:localPerformanceNumber'),

    # extra
    'comment': strcont('TEMP:providerNote'),
    'note': strcont('TEMP:providerNote'),
}


class SubjectExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Subject', 'owl:NamedIndividual']},
        'required': ['subject_id', 'species'],
        'properties': {
            'subject_id': {'type': 'string',
                           #'context_runtime':
                           #lambda base: {'@id': '@id',
                                         #'@type': '@id',
                                         #'@context': {'@base': f'{base}subjects/'}}
                           },
            'ear_tag_number': strcont('TEMP:localIdAlt'),
            'treatment': strcont('TEMP:hadExperimentalTreatmentApplied'),
            'initial_weight': strcont('sparc:animalSubjectHasWeight'),
            'height_inches': strcont('TEMP:subjectHasHeight'),
            'gender': strcont('sparc:hasGender'),
            'body_mass_weight': strcont('TEMP:subjectHasWeight'),

                        'anesthesia': strcont('TEMP:wasAdministeredAnesthesia'),
            'stimulation_site': strcont('sparc:spatialLocationOfModulator'),  # TODO ontology
            'stimulator': strcont('sparc:stimulatorUtilized'),
            'stimulating_electrode_type': strcont('sparc:stimulatorUtilized'),  # FIXME instance vs type

            'nerve': strcont('TEMPRAW:involvesAnatomicalRegion'),  # FIXME can be more specific than this
            **subsam_common_properties
        },}


class SubjectsExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        #'context_value': {'@id': 'graph-subjects', CONTEXT_ROOT: None},  # not needed, schemas have their own context
        'required': ['subjects'],
        'properties': {
            'subjects': {
                'type': 'array',
                #'context_runtime': lambda base: {'@id': f'{base}subjects',},
                'minItems': 1,
                # FIXME there is currently no way to require that
                # a key be unique among all objects
                'items': SubjectExportSchema.schema,},
            'errors': ErrorSchema.schema,
            'software': _software_schema}}

    # FIXME not implemented
    extras = [['unique', ['subjects', '*', 'subject_id']]]


class SubjectsSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(SubjectsExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class SampleExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:Sample', 'owl:NamedIndividual']},
        'required': ['sample_id', 'subject_id'],
        'properties': {
            'sample_id': {'type': 'string',},
            'subject_id': {'type': 'string',},
            'primary_key': {'type': 'string',
                            'context_runtime': [  # TODO -> derive after add ?
                                '#/meta/uri_api',
                                # FIXME the @id mapping seems to break the @context !?
                                (lambda uri_api: {
                                    '@id': '@id',
                                    '@context': {
                                        '@base': uri_api + '/samples/'}}),
                                '#/samples/@context/primary_key'],
                            #'context_runtime':
                            #lambda base: {'@id': '@id',
                                          #'@type': '@id',
                                          #'@context': {'@base': f'{base}samples/'}}
                            },
            'specimen': {'type': 'string',
                         },  # what are we expecting here?
            'specimen_anatomical_location': {'type': 'string',
                                             },

                        'disease': {'type': 'string',
                                    },
            'reference_atlas': {'type': 'string',
                                },
            'protocol_title': {'type': 'string',
                               },
            'protocol_url_or_doi': _protocol_url_or_doi_schema 
            ,
            'experimental_log_file_name': {'type': 'string',
                                           },
            **subsam_common_properties
        },}


class SamplesFileExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'required': ['samples'],
        'properties': {
            'samples': {
                'type': 'array',
                #'context_runtime': lambda base: {'@id': f'{base}samples',},
                'minItems': 1,
                'items': SampleExportSchema.schema,},
            'errors': ErrorSchema.schema,
            'software': _software_schema}}

    # FIXME not implemented
    extras = [['unique', ['samples', '*', 'sample_id']]]


class SamplesFileSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(SamplesFileExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class ManifestRecordExportSchema(JSONSchema):
    schema = {
        'allOf': [
            {'type': 'object',
             'properties': {
                 'pattern': {
                     'type': 'string',
                 },
                 'filename': {
                     'type': 'string',
                 },
                 'timestamp': {
                     'type': 'string',
                     'pattern': iso8601pattern,
                 },
                 'description': {
                     'type': 'string',
                 },
                 'file_type': {
                     'type': 'string',
                 },
                 'data_type': {
                     'type': 'string',
                 },
             },
             },
            {'oneOf': [
                {'required': ['pattern']},
                {'required': ['filename', 'timestamp']},
            ]},
        ]
    }


class ManifestFileExportSchema(JSONSchema):  # FIXME TODO FileObjectSchema ??
    schema = {
        'type': 'object',
        'jsonld_include': {'@type': ['sparc:File', 'owl:NamedIndividual']},
        'required': ['dataset_relative_path', 'manifest_records'],  # TODO uri_api
        'properties': {
            'dataset_relative_path': {'type': 'string'},
            'manifest_records': {
                'type': 'array',
                'minItems': 1,
                'items': ManifestRecordExportSchema.schema,
            },
            'errors': ErrorSchema.schema,
        }
    }


ManifestFileSchema = ManifestFileExportSchema


class ManifestFilesExportSchema(JSONSchema):
    schema = {
        'type': 'object',
        'required': ['manifest_files'],
        'properties': {
            'manifest_files': {
                'type': 'array',
                'minItems': 1,
                'items': ManifestFileExportSchema.schema,
            },
            'errors': ErrorSchema.schema,
        }
    }


class MetaOutExportSchema(JSONSchema):
    __schema = copy.deepcopy(DatasetDescriptionExportSchema.schema)
    extra_required = ['award_number',
                      'principal_investigator',
                      'species',
                      'organ',
                      'modality',
                      'techniques',
                      'contributor_count',
                      'uri_human',  # from DatasetMetadata
                      'uri_api',  # from DatasetMetadata
                      'files',
                      'dirs',
                      'size',
                      'folder_name',  # from DatasetMetadat
                      'title',
                      'schema_version',
                      'number_of_subjects',
                      'number_of_samples',
                      'timestamp_created',
                      'timestamp_updated',
                      'timestamp_updated_contents',
                      #'subject_count',
                      #'sample_count',
    ]
    __schema['required'].remove('contributors')
    __schema['required'].remove('name')
    __schema['required'] += extra_required
    __schema['properties'].pop('contributors')
    __schema['properties'].pop('name')
    __schema['properties'].update({
        'errors': ErrorSchema.schema,
        'dirs': {'type': 'integer'},
        'doi': EIS._allOf(DoiSchema, context_value='TEMP:hasDoi'),
        'files': {'type': 'integer',
                  'context_value': inttype('TEMP:numberOfFiles'),
                  },
        'size': {'type': 'integer',
                 'context_value': inttype('TEMP:hasSizeInBytes'),
                 },
        'folder_name': {'context_value': 'rdfs:label',
                        **string_noltws},
        'title': {'context_value': 'dc:title',
                  **string_noltws},
        'timestamp_created': {'type': 'string',
                              'pattern': iso8601pattern,
                              'context_value': 'TEMP:wasCreatedAtTime',
                              },
        'timestamp_updated': {'type': 'string',
                              'pattern': iso8601pattern,
                              'context_value': 'TEMP:wasUpdatedAtTime',
                              },
        'timestamp_updated_contents': {'type': 'string',
                                       'pattern': iso8601pattern,
                                       'context_value': 'TEMP:contentsWereUpdatedAtTime',
                                       },
        'uri_human': {'type': 'string',
                      'pattern': r'^https://app\.blackfynn\.io/N:organization:',  # FIXME proper regex
                      'context_value': idtype('TEMP:hasUriHuman'),
        },
        'uri_api': {'type': 'string',
                    'pattern': r'^https://api\.blackfynn\.io/(datasets|packages)/',  # FIXME proper regex
                    #'context_value': idtype('TEMP:hasUriApi'),
                    'context_value': idtype('@id'),
        },
        'award_number': {'type': 'string',
                         'pattern': '^OT|^U18',
                         'context_value': idtype('TEMP:hasAwardNumber', base=TEMP['awards/']),
                         },
        'principal_investigator': {'type': 'string',
                                   'context_value': idtype('TEMP:hasResponsiblePrincipalInvestigator'),
                                   },
        'protocol_url_or_doi': {'type': 'array',
                                'minItems': 1,
                                'items': _protocol_url_or_doi_schema,
                                'context_value': idtype('TEMP:hasProtocol'),
                                },
        'additional_links': {'type': 'array',
                             'minItems': 1,
                             'items': {'type': 'string'},
                             'context_value': idtype('TEMP:hasAdditionalLinks'),  # TODO
                             },
        'species': {'type': 'string',
                    'context_value': idtype('isAbout'),
                    },
        'schema_version': {'type': 'string'},
        'organ': {'type': 'array',
                  'minItems': 1,
                  'items': EIS._allOf(OntTermSchema),
                  'context_value': idtype('isAbout'),
                  },
        'modality': {'type': 'array',
                     'minItems': 1,
                     'items': {
                         'type': 'string',
                     },
                    'context_value': idtype('TEMP:hasExperimentalModality'),
        },
        'techniques': {'type': 'array',
                       'minItems': 1,
                       'items': EIS._allOf(OntTermSchema),
                       # FIXME usedProtocolThatEmployedTechniques ...
                       'context_value': idtype('TEMP:protocolEmploysTechnique'),
        },

        # TODO $ref these w/ pointer?
        # in contributor etc?

        'subject_count': {'type': 'integer',
                          'context_value': inttype('TEMP:hasNumberOfSubjects'),
                          },
        'sample_count': {'type': 'integer',
                         'context_value': inttype('TEMP:hasNumberOfSamples'),
                         },
        'contributor_count': {'type': 'integer',
                              'context_value': inttype('TEMP:hasNumberOfSamples'),
                              },})

    schema = {'allOf': [__schema,
                        {'anyOf': [
                            #{'required': ['number_of_subjects']},
                            #{'required': ['number_of_samples']},
                            {'required': ['subject_count']},  # FIXME extract subjects from samples ?
                            {'required': ['sample_count']}
                        ]}]}


class MetaOutSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(MetaOutExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class DatasetOutExportSchema(JSONSchema):
    """ Schema that adds the id field since investigators shouldn't need to know
        the id to check the integrity of their data. We need it because that is
        a key piece of information that we use to link everything together. """

    __schema = copy.deepcopy(DatasetStructureSchema.schema['allOf'][0])
    __schema['required'] = ['id',
                            'meta',
                            'contributors',
                            'prov']
    __schema['properties'] = {
        'id': {'type': 'string',  # ye old multiple meta/bf id issue
               'pattern': '^N:dataset:'},
        'meta': MetaOutExportSchema.schema,
        'prov': ProvSchema.schema,
        'errors': ErrorSchema.schema,
        'contributors': ContributorsOutExportSchema.schema,
        'creators': CreatorsExportSchema.schema,
        # FIXME subjects_file might make the name overlap clearer
        'subjects': SubjectsExportSchema.schema['properties']['subjects'],  # FIXME SubjectsOutSchema
        'samples': SamplesFileExportSchema.schema['properties']['samples'],
        # 'identifier_metadata': {'type': 'array'},  # FIXME temporary
        'resources': {'type': 'array',
                      'items': {'type': 'object'},},
        'submission': {'type': 'object',},
        'inputs': {'type': 'object',
                   # TODO do we need errors at this level?
                   'properties': {'dataset_description_file': DatasetDescriptionSchema.schema,
                                  'submission_file': SubmissionSchema.schema,
                                  'subjects_file': SubjectsSchema.schema,
                                  'samples_file':SamplesFileSchema.schema,
                                  },},}

    # FIXME switch to make samples optional since subject_id will always be there even in samples?
    schema = {
        'jsonld_include': {'@type': ['sparc:Dataset', 'owl:NamedIndividual']},
        'allOf': [__schema,
                  {'anyOf': [
                      {'required': ['subjects']},  # FIXME extract subjects from samples ?
                      {'required': ['samples']}
                  ]}]}


class DatasetOutSchema(JSONSchema):
    context = lambda : ({}, None)
    __schema = copy.deepcopy(DatasetOutExportSchema.schema)
    schema = JApplyRecursive(EIS._to_pattern, __schema)


class StatusSchema(JSONSchema):
    schema = {'type': 'object',
              'required': ['submission_index', 'curation_index', 'error_index',
                           'submission_errors', 'curation_errors'],
              'properties': {
                  'status_on_platform': {'type': 'string',
                                         'context_value': 'TEMP:blackfynnPlatformStatus',
                                         },
                  'submission_index': {'type': 'integer', 'minimum': 0,
                                       'context_value': inttype('TEMP:submissionIndex'),
                                       },
                  'curation_index': {'type': 'integer', 'minimum': 0,
                                     'context_value': inttype('TEMP:curationIndex'),
                                     },
                  'error_index': {'type': 'integer', 'minimum': 0,
                                  'context_value': inttype('TEMP:errorIndex'),
                                  },
                  'submission_errors':{'type':'array',  # allow these to be empty
                                       'items': {'type': 'object'},
                                       },
                  'curation_errors':{'type':'array',  # allow these to be empty
                                     'items': {'type': 'object'},
                                     },
              }}


class PostSchema(JSONSchema):
    """ This is used to validate only the status schema part that is added to the DatasetOutSchema """
    schema = {
        'type': 'object',
        'required': ['status'],
        'properties': {'status': StatusSchema.schema,}
    }


class SummarySchema(JSONSchema):
    # TODO add expected counts
    schema = {'type': 'object',
              'required': ['id', 'meta', 'prov'],
              'properties': {'id': {'type': 'string',
                                    'pattern': '^N:organization:'},
                             'meta': {'type': 'object',
                                      'required': ['folder_name', 'count', 'uri_api', 'uri_human'],
                                      'properties': {'folder_name': {'type': 'string'},
                                                     # FIXME common source for these with MetaOutSchema
                                                     'uri_human': {'type': 'string',
                                                                   'pattern': r'^https://app\.blackfynn\.io/N:organization:',
        },
                                                     'uri_api': {'type': 'string',
                                                                 'pattern': r'^https://api\.blackfynn\.io/organizations/N:organization:',
                                                     },
                                                     'count': {'type': 'integer'},},},
                             'prov': {'type': 'object'},
                             'datasets': {'type': 'array',
                                          'minItems': 1,
                                          'items': {'type': 'object'},  # DatasetOutSchema already validated
                             }}}


class HeaderSchema(JSONSchema):
    schema = {'type': 'array',
              'items': {'type': 'string'}}


# constraints on MIS classes
class MISSpecimenSchema(JSONSchema):
    schema = {
        'type': 'object',
        'properties': {
            'TEMP:hasAge': {},
            'TEMP:hasAgeCategory': {},  # hasDevelopmentalStage
            'TEMP:hasBiologicalSex': {},  # TEMP because sparc:hasPhenotypicSex
            'TEMP:hasWeight': {},
            'TEMP:hasGenus': {},
            'TEMP:hasSpecies': {},
            'TEMP:hasStrain': {},  # race/ethnicity?
            'TEMP:hasRRID': {},
            'TEMP:hasGeneticVariation': {},

            'sparc:hasBodyMassIndex': {} ,
            'sparc:hasClinicalInformation': {},
            'sparc:hasDiseaseInformation': {},
            'sparc:hasDrugInformation': {},
            'sparc:hasFamilyHistory': {},
            'sparc:hasGender': {},
            'TEMP:hasHandedness': {},
            'TEMP:hasRace': {},
            'TEMP:hasEthnicity': {},
        },
        'oneOf': [
            {'allOf': [
                {'allOf':
                 [{'required': [
                     'TEMP:hasSpecies',
                     'TEMP:hasBiologicalSex',
                     'TEMP:hasWeight',]},
                  {'oneOf': [{'required': ['TEMP:hasAge']},
                             {'required': ['TEMP:hasAgeCategory']}]}]},
                {'oneOf': [
                    {'properties': {'TEMP:hasSpecies': {'enum': ['NCBITaxon:9606']}},
                     'required': [
                         'sparc:hasBodyMassIndex',
                         'sparc:hasClinicalInformation',
                         'sparc:hasDiseaseInformation',
                         'sparc:hasDrugInformation',
                         'sparc:hasFamilyHistory',
                         #'sparc:hasGender',
                         #'TEMP:hasHandedness',
                         'TEMP:hasRace',
                         'TEMP:hasEthnicity',
                     ]},
                    {'allOf': [
                        {'not': {'properties': {'TEMP:hasSpecies': {'enum': ['NCBITaxon:9606']}}}},
                        {'required': [
                            #'TEMP:hasSupplier',
                            #'TEMP:hasStrain',  # RRID? race/ethnicity?
                            'TEMP:hasRRID',
                         ]},
                    ]},
                ]},
            ]},
        ],
    }


class MISDatasetSchema(JSONSchema):
    moves = [  # FIXME this goes somehwere else
        # from datasetoutschema -> mis
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [[''], ['']],
        [['resources'], ['']],
        [['meta', 'protocol_url_or_doi'], ['']],
        #[['meta'], ['sparc:Resource']],
        [['subjects'], ['sparc:Specimen']],
        [['contributors'], ['sparc:Researcher']],  # ah the role confusion
    ]
    schema = {'type': 'object',
              'required': ['curies', 'rdf:type'],
              'properties': {
                  'curies': {'type': 'object'},
                  'rdf:type': {'type': 'string',
                               'pattern': 'sparc:Resource'},
                  #'sparc:'
                  'sparc:Specimen': {'type': 'array',
                                     'items': MISSpecimenSchema.schema},
              }
              }



import re
import sys
import json
import types
import pprint
from itertools import chain
from collections import Counter, defaultdict
import idlib
import htmlfn as hfn
import augpathlib as aug
from hyputils import hypothesis as hyp
from protcur import document as ptcdoc
from pyontutils.core import OntGraph, OntRes, OntResIri, OntResPath
from pyontutils.utils import UTCNOWISO, anyMembers
from sparcur import sheets
from sparcur.core import (OntId,
                          OntTerm,
                          get_all_errors,
                          adops,
                          register_type,
                          fromJson)
from sparcur.utils import want_prefixes, log as _log
from sparcur.paths import Path
from sparcur.config import auth
from sparcur.curation import ExporterSummarizer
from sparcur.curation import Integrator, DatasetObject  # FIXME

log = _log.getChild('reports')


def normalize_exact(e, *, lower=False):
    if e is not None:
        e = e.replace('\n', ' ')  # sigh
        e = e.replace('\t', ' ')
        e = e.replace('\xA0', ' ')  # nbsp
        e = e.strip().rstrip()
        if lower:
            e = e.lower()
        return e

    return ''


def normtext(t, *, lower=False):
    if 'https://' in t:
        return ''
    else:
        t = t.strip().rstrip()
        if lower:
            t = t.lower()

        return t


class SparqlQueries:
    """ Creates SPARQL query templates. """

    def __init__(self, nsm=None):
        from rdflib.plugins import sparql
        self.sparql = sparql

        self.nsm = nsm if nsm else OntGraph().namespace_manager
        self.prefixes = dict(self.nsm)

    def dataset_about(self):
        # FIXME this will return any resource matching isAbout:
        query = """
            SELECT ?dataset
            WHERE {
                ?dataset rdf:type sparc:Resource .
                ?dataset isAbout: ?about .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_subjects(self) -> str:
        """ Get all subject groups and dataset associated with subject input.

        :returns: list of tuples containing: subject, subjects group, and subjects dataset.
        """
        query = """
            SELECT ?dataset ?subj
            WHERE {
                ?startsubj TEMP:hasDerivedInformationAsParticipant ?dataset .
                ?subj  TEMP:hasDerivedInformationAsParticipant ?dataset .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_groups(self) -> str:
        """ Get all subject groups and dataset associated with subject input.

        :returns: list of tuples containing: subject, subjects group, and subjects dataset.
        """
        query = """
            SELECT ?dataset ?group ?subj
            WHERE {
                ?startsubj TEMP:hasDerivedInformationAsParticipant ?dataset .
                ?subj  TEMP:hasDerivedInformationAsParticipant ?dataset .
                ?subj  TEMP:hasAssignedGroup ?group .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_bundle(self) -> str:
        """ Get all related datasets of subject.

        :returns: list of tuples containing: subject & subjects shared dataset.
        """
        query = """
            SELECT ?dataset
            WHERE {
                ?startdataset TEMP:collectionTitle ?string .
                ?dataset  TEMP:collectionTitle ?string .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_subject_species(self):
        # FIXME how to correctly init bindings to multiple values ...
        query = """
            SELECT DISTINCT ?dataset
            WHERE {
                VALUES ?species { "human" "homo sapiens" } .
                ?dataset TEMP:isAboutParticipant ?subject .
                ?subject sparc:animalSubjectIsOfSpecies ?species .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def award_affiliations(self):
        query = """
            SELECT DISTINCT ?award ?affiliation
            WHERE {
                ?dataset TEMP:hasAwardNumber ?award .
                ?contributor TEMP:contributorTo ?dataset .
                ?contributor TEMP:hasAffiliation ?affiliation .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def dataset_milestone_completion_date(self):
        query = """
            SELECT DISTINCT ?dataset ?date
            WHERE {
                ?dataset TEMP:milestoneCompletionDate ?date .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def protocol_techniques(self):
        query = """
            SELECT DISTINCT ?protocol ?technique
            WHERE {
                ?protocol rdf:type sparc:Protocol .
                ?protocol TEMP:protocolEmploysTechinque ?technique .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)

    def protocol_aspects(self):
        query = """
            SELECT DISTINCT ?protocol ?aspect
            WHERE {
                ?protocol rdf:type sparc:Protocol .
                ?protocol TEMP:protocolInvolvesAspect ?aspect .
            }
        """
        return self.sparql.prepareQuery(query, initNs=self.prefixes)



class TtlFile:

    def __init__(self, ontres, ontres_compare_to=None):
        import rdflib
        from pyontutils import namespaces as ns
        self.rdflib = rdflib
        self.ns = ns

        self.ontres = ontres

        if ontres_compare_to is None:
            self.ontres_compare_to = ontres
        else:
            self.ontres_compare_to = ontres_compare_to

        self.graph = self.ontres.graph
        self.graph_compare_to = self.ontres_compare_to.graph

        self.queries = SparqlQueries(nsm=self.graph.namespace_manager)

    def _to_human(self, row):
        return [self.graph.namespace_manager.qname(cell)
                if isinstance(cell, rdflib.URIRef) else
                cell.toPython() for cell in row]

    def changes(self):
        added, removed, changed = self.graph.subjectsChanges(self.graph_compare_to)

    def milestones(self):
        query = self.queries.dataset_milestone_completion_date()
        res = list(self.graph.query(query))
        rows = sorted(self._to_human(row) for row in res)
        header = [['dataset', 'milestone comp date']]
        return header + rows
        return self._print_table(header + rows,
                                 title=f'Dataset milestone completion dates')

    def mis(self, ext=None):
        rdflib = self.rdflib
        ns = self.ns
        ontres = self.ontres

        counts = Counter(ontres.graph.predicates())
        g = OntGraph(namespace_manager=ontres.graph.namespace_manager)

        [(g.add((s, ns.ilxtr.numberOfOccurrences, rdflib.Literal(count))),
          g.add((s, ns.rdf.type, ns.ilxtr.MISPredicate)))
         for s, count in counts.items()
         if g.namespace_manager.qname(s).split(':')[0]
         not in ('rdf', 'rdfs', 'owl')]
        g.debug()

        rows = [[g.namespace_manager.qname(s), o.toPython()]
                for s, o in g[:ns.ilxtr.numberOfOccurrences:]]
        header = [['curie', 'count']]
        return header + sorted(rows, key=lambda r:-r[-1])  # FIXME only Report has access to -C :/


class Report:

    @property
    def _sort_key(self):
        if self.options.sort_count_desc:
            return lambda kv: -kv[-1]
        else:
            return lambda kv: kv

    def all(self):
        #self.access()  # must be on live data
        # TODO add a prov report for when this was last run etc.
        self.contributors()
        self.filetypes()
        self.samples()
        self.subjects()
        self.completeness()
        self.keywords()
        #self.size(dirs=list(self.datasets_local))  # TODO needs to be reworked to use latest_ir
        self.overview()
        #self.test()  # just a test
        #self.errors()  # TODO too much for now
        #self.pathids()  # too big/slow
        if hasattr(self, 'anchor'):  # FIXME hack
            # branch used to detect if we have a chance of finding the xml export
            # xml ir is not exported with all the rest at the moment
            self.mbf()

        self.terms()
        self.mis()
        for tag in (
                'protc:input',  # this is slower now
                'protc:input-instance',
                'protc:implied-input',
                'protc:aspect',
                'protc:implied-aspect',
                'protc:executor-verb',
                'protc:parameter*',
                'protc:invariant',
                'protc:output',):
            # FIXME hack around broken anno_tags impl
            self.anno_tags(tag)

    def access(self, ext=None):
        """ Report on datasets that are in the master sheet but which the
            automated pipelines do not have access to. """
        if self.anchor.id == 'N:organization:fake-organization-id':
            # FIXME need a way to generalize/fix this so that organization
            # can vary ...
            print('Not in correct organization for running access report.')
            sys.exit(9999)

        if self.options.server:
            def fmt(d): return hfn.atag(d.uri_human, d.id, new_tab=True)
        else:
            def fmt(d): return d.uri_human

        if self.options.server:
            uri_human = sheets.Master._uri_human()
            def fmtmc(s): return hfn.atag(uri_human, s, new_tab=True)
        else:
            def fmtmc(s): return s

        o = sheets.Overview()
        remote = self.anchor.remote
        remote_datasets = remote.children

        master_sheet_ids = o.dataset_ids()
        bf_api_ids = set(d.id for d in remote_datasets)

        missing_from_api = sorted(master_sheet_ids - bf_api_ids)
        missing_datasets = [self.BlackfynnRemote(i, local_only=True)
                            for i in missing_from_api]
        missing_uris = [fmt(d) for d in missing_datasets]
        rows = [['', ''],
                [fmtmc('Master Count'), len(master_sheet_ids)],
                ['BF API Count', len(bf_api_ids)],
                ['No API Count', len(missing_from_api)],
                *[[i, u] for i, u in zip(missing_from_api, missing_uris)]
        ]
        return self._print_table(rows, title='Access Report', ext=ext)

    _pyru_loaded = False
    def _data_ir(self):
        if self.options.raw:
            data = self.summary.data_for_export(UTCNOWISO())
        elif self.options.export_file:
            if not self.__class__._pyru_loaded:
                self.__class__._pyru_loaded = True
                from pysercomb.pyr import units as pyru
                [register_type(c, c.tag) for c in (pyru._Quant, pyru.Range)]

            with open(self.options.export_file, 'rt') as f:
                data = fromJson(json.load(f))
        else:
            from sparcur import export as ex
            data = self._export(ex.Export).latest_ir
        return data

    @sheets.Reports.makeReportSheet('id')
    def contributors(self, ext=None):
        data = self._data_ir()

        datasets = data['datasets']
        unique = {c['id']:c for d in datasets
                  if 'contributors' in d
                  for c in d['contributors']}
        contribs = sorted(unique.values(),
                          key=lambda c: (c['last_name']  # pYthON dOEsNt hAEV mUlTiLiNE lAmBDaS
                                         if 'last_name' in c else
                                         (c['contributor_name'] if
                                          'contributor_name' in c else
                                          'zzzzzzzzzzzzzzzzzzzzzzzzzzzz')))
        #contribs = sorted((dict(c) for c in
                           #set(frozenset({k:tuple(v) if isinstance(v, list) else
                                          #(frozenset(v.items()) if isinstance(v, dict) else v)
                                          #for k, v in c.items()}.items())
                               #for d in datasets
                               #if 'contributors' in d
                               #for c in d['contributors']
                               #if not log.info(lj(c)))),
                          #key=lambda c: c['last_name'] if 'last_name' in c else c['name'])

        if self.options.debug:
            breakpoint()

        rows = [['id', 'last', 'first', 'PI', 'No Orcid']] + [[
            c['id'],
            c['last_name'] if 'last_name' in c else '',
            c['first_name'] if 'first_name' in c else '',
            'x' if 'contributor_role' in c and 'PrincipalInvestigator' in c['contributor_role'] else '',
            'x' if 'orcid' not in c['id'] else '']
            for c in contribs]

        return self._print_table(rows,
                                 title='Contributors Report',
                                 ext=ext)

    def tofetch(self, dirs=None, ext=None):
        if dirs is None:
            dirs = self.options.directory
            if not dirs:
                dirs.append(self.cwd.as_posix())

        data = []

        def dead(p):
            raise ValueError(p)

        for d in dirs:
            if not Path(d).is_dir():
                continue  # helper files at the top level, and the symlinks that destory python
            path = Path(d).resolve()
            paths = path.rchildren #list(path.rglob('*'))
            path_meta = {p:p.cache.meta if p.cache else dead(p) for p in paths
                         if p.suffix not in ('.swp',)}
            outstanding = 0
            total = 0
            tf = 0
            ff = 0
            td = 0
            uncertain = False  # TODO
            for p, m in path_meta.items():
                #if p.is_file() and not any(p.stem.startswith(pf) for pf in self.spcignore):
                if p.is_file() or p.is_broken_symlink():
                    s = m.size
                    if s is None:
                        uncertain = True
                        continue

                    tf += 1
                    if s:
                        total += s

                    #if '.fake' in p.suffixes:
                    if p.is_broken_symlink():
                        ff += 1
                        if s:
                            outstanding += s

                elif p.is_dir():
                    td += 1

            data.append([path.name,
                         aug.FileSize(total - outstanding),
                         aug.FileSize(outstanding),
                         aug.FileSize(total),
                         uncertain,
                         (tf - ff),
                         ff,
                         tf,
                         td])

        formatted = [[n, l.hr, o.hr, t.hr if not u else '??', lf, ff, tf, td]
                     for n, l, o, t, u, lf, ff, tf, td in
                     sorted(data, key=lambda r: (r[4], -r[3]))]
        rows = [['Folder', 'Local', 'To Retrieve', 'Total', 'L', 'R', 'T', 'TD'],
                *formatted]

        return self._print_table(rows, title='File size counts', ext=ext)

    # TODO generator issue
    #@sheets.Reports.makeReportSheet('suffix', 'mimetype', 'magic_mimetype')
    def filetypes(self, ext=None):
        key = self._sort_key
        paths = self.paths if self.paths else (self.cwd,)
        paths = [c for p in paths for c in p.rchildren if not c.is_dir()]
        rex = re.compile('^\.[0-9][0-9][0-9A-Z]$')
        rex_paths = [p for p in paths if re.match(rex, p.suffix)]
        paths = [p for p in paths if not re.match(rex, p.suffix)]

        def count(thing):
            return sorted([(k if k else '', v) for k, v in
                            Counter([getattr(f, thing)
                                     for f in paths]).items()], key=key)

        each = {t:count(t) for t in ('suffix', 'mimetype', '_magic_mimetype')}
        each['suffix'].append((rex.pattern, len(rex_paths)))

        for title, rows in each.items():
            yield self._print_table(((title, 'count'), *rows),
                                    title=title.replace('_', ' ').strip())

        all_counts = sorted([(*[m if m else '' for m in k], v) for k, v in
                             Counter([(f.suffix, f.mimetype, f._magic_mimetype)
                                      for f in paths]).items()], key=key)

        header = ['suffix', 'mimetype', 'magic mimetype', 'count']
        return self._print_table((header, *all_counts),
                                 title='All types aligned (has duplicates)',
                                 ext=ext)

    @sheets.Reports.makeReportSheet('column_name')
    def samples(self, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        key = self._sort_key
        # FIXME we need the blob wrapper in addition to the blob generator
        # FIXME these are the normalized ones ...
        samples_headers = tuple(k for dataset_blob in datasets
                                 if 'samples' in dataset_blob  # FIXME inputs?
                                 for samples_blob in dataset_blob['samples']
                                 for k in samples_blob)
        counts = tuple(kv for kv in sorted(Counter(samples_headers).items(),
                                            key=key))

        index_col_name = 'Column Name'
        if ext is None:
            index_col_name += f' unique = {len(counts)}'
        rows = ((index_col_name, '#'), *counts)
        return self._print_table(rows,
                                 title='Samples Report',
                                 ext=ext)

    @sheets.Reports.makeReportSheet('column_name')
    def subjects(self, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        key = self._sort_key
        # FIXME we need the blob wrapper in addition to the blob generator
        # FIXME these are the normalized ones ...
        subjects_headers = tuple(k for dataset_blob in datasets
                                 if 'subjects' in dataset_blob  # FIXME inputs?
                                 for subject_blob in dataset_blob['subjects']
                                 for k in subject_blob)
        counts = tuple(kv for kv in sorted(Counter(subjects_headers).items(),
                                            key=key))

        index_col_name = 'Column Name'
        if ext is None:
            index_col_name += f' unique = {len(counts)}'
        rows = ((index_col_name, '#'), *counts)
        return self._print_table(rows,
                                 title='Subjects Report',
                                 ext=ext)

    @sheets.Reports.makeReportSheet('id')  # FIXME vs path
    def completeness(self, ext=None):
        if self.options.raw:
            raw = self.summary.completeness
        else:
            from sparcur import export as ex
            datasets = self._data_ir()['datasets']
            raw = [ExporterSummarizer._completeness(data) for data in datasets]

        def rformat(i, si, ci, ei, name, id, award, organ):
            if self.options.server and isinstance(ext, types.FunctionType):
                # rsurl = 'https://projectreporter.nih.gov/reporter_searchresults.cfm'
                dataset_dash_url = self.url_for('route_datasets_id', id=id)
                errors_url = self.url_for('route_reports_errors_id', id=id)
                si = hfn.atag(errors_url + '#submission', si)
                ci = hfn.atag(errors_url + '#curation', ci)
                ei = hfn.atag(errors_url + '#total', ei)
                name = hfn.atag(dataset_dash_url, name)
                id = hfn.atag(dataset_dash_url, id[:10] + '...')
                award = (
                    hfn.atag(('https://scicrunch.org/scicrunch/data/source/'
                              f'nif-0000-10319-1/search?q={award}'), award)
                    if award else 'MISSING')
                organ = organ if organ else ''
                if isinstance(organ, list) or isinstance(organ, tuple):
                    organ = ' '.join([o.atag() for o in organ])
                if isinstance(organ, OntTerm):
                    organ = organ.atag()
            else:
                award = award if award else ''
                organ = ((repr(organ) if isinstance(organ, OntTerm) else organ)
                         if organ else '')
                if isinstance(organ, list):
                    organ = ' '.join([repr(o) for o in organ])

            return (i + 1, si, ci, ei, name, id, award, organ)

        rows = [('', 'SI', 'CI', 'EI', 'name', 'id', 'award', 'organ')]
        rows += [rformat(i, *cols) for i, cols in enumerate(sorted(
            raw, key=lambda t: (t[0], t[1], t[5] if t[5] else 'z' * 999, t[3])))]

        return self._print_table(rows,
                                 title='Completeness Report',
                                 ext=ext)

    @sheets.Reports.makeReportSheet()
    def keywords(self, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        _rows = [sorted(set(dataset_blob.get('meta', {}).get('keywords', [])),
                        key=lambda v: -len(v))
                    for dataset_blob in datasets]
        rows = [list(r) for r in sorted(
            set(tuple(r) for r in _rows if r),
            key = lambda r: (len(r), tuple(len(c) for c in r if c), r))]
        header = [[f'{i + 1}' for i, _ in enumerate(rows[-1])]]
        rows = header + rows
        return self._print_table(rows,
                                 title='Keywords Report',
                                 ext=ext)


    @sheets.Reports.makeReportSheet('id')
    def size(self, dirs=None, ext=None):
        if dirs is None:
            dirs = self.paths
        intrs = [Integrator(p) for p in dirs]
        if not intrs:
            intrs = self.cwdintr,

        rows = [['path', 'id', 'sparse', 'dirs', 'files', 'size', 'hr'],
                *sorted([[d.name,
                          d.id,
                          'x' if d.path.cache.is_sparse() else '',
                          c['dirs'], c['files'], c['size'], c['size'].hr]
                         for d in intrs
                         for c in (d.datasetdata.counts,)],
                        key=lambda r: -r[-2])]

        return self._print_table(rows, title='Size Report',
                                 align=['l', 'l', 'r', 'r', 'r', 'r', 'r'],
                                 ext=ext)

    @sheets.Reports.makeReportSheet('id')
    def overview(self, ext=None):
        data = self._data_ir()
        datasets = data['datasets']
        hrm = {'award': ['meta', 'award_number'],
               'id': ['id'],
               'name': ['meta', 'folder_name'],
               'errors': ['status', 'error_index'],
               'updated': ['status', 'updated'],
               'published': ['meta', 'doi'],
               'milestone_completion_date': ['submission',
                                             'milestone_completion_date'],
               # TODO
               # subject_count
               # sample_count
               # contributor_count
               }
        header = tuple(hrm)  # TODO counts ?
        paths = tuple(hrm.values())
        def getl(d):
            return [adops.get(d, path, on_failure='') for path in paths]
        _rows = sorted([getl(d) for d in datasets],
                       key=lambda r: (r[-2].asStr()
                                      if isinstance(r[-2], idlib.Stream) else
                                      r[-2],
                                      print(r),
                                      r[0],
                                      r[-3],
                                      r[-4],
                                      ))
        rows = [header] + _rows
        return self._print_table(rows, title='Dataset Report',
                                 align=['l', 'l', 'l', 'r', 'r', 'r', 'r'],
                                 ext=ext)

    def test(self, ext=None):
        rows = [['hello', 'world'], [1, 2]]
        return self._print_table(rows, title='Report Test', ext=ext)


    #@sheets.Reports.makeReportSheet('id')  # TODO bad return format right now
    def errors(self, *, id=None, ext=None):
        data = self._data_ir()
        datasets = data['datasets']

        if self.cwd != self.anchor:
            id = self.cwd.cache.dataset.id

        if id is not None:
            if not id.startswith('N:dataset:'):
                return []

            def pt(rendered_table, title=None):
                """ passthrough """
                return rendered_table

            import htmlfn as hfn
            for dataset_blob in datasets:
                if dataset_blob['id'] == id:
                    dso = DatasetObject.from_json(dataset_blob)
                    title = f'Errors for {id}'
                    urih = dataset_blob['meta']['uri_human']
                    formatted_title = (
                        hfn.h2tag(f'Errors for {hfn.atag(urih, id)}<br>\n') +
                        (hfn.h3tag(dataset_blob['meta']['title']
                                   if 'title' in dataset_blob['meta'] else
                                   dataset_blob['meta']['folder_name'])))
                    log.info(list(dataset_blob.keys()))
                    errors = list(dso.errors)
                    return [(self._print_table(e.as_table(), ext=pt))
                            for e in errors], formatted_title, title
        else:
            pprint.pprint(
                sorted([(d['meta']['folder_name'],
                         [e['message'] for e in get_all_errors(d)])
                        for d in datasets],
                       key=lambda ab: -len(ab[-1])))

    def pathids(self, ext=None):
        base = self.project_path.parent
        rows = [['path', 'id']] + sorted([c.relative_to(base), c.cache.id]#, c.cache.uri_api, c.cache.uri_human]
                                         # slower to include the uris
                                         for c in chain((self.cwd,),
                                                        self.cwd.rchildren)
        )
        return self._print_table(rows, title='Path identifiers', ext=ext)

    #@sheets.Reports.makeReportSheet('id')  # TODO bad return format right now
    def mbf(self, ext=None):
        et = tuple()
        from sparcur.extract import xml as exml

        def settype(mimetype):
            return {exml.ExtractMBF.mimetype: 'MBF Metadata',
                    exml.ExtractNeurolucida.mimetype: 'Neurolucida',}[mimetype]

        if self.options.raw:
            blob_ir = self.parent.export()
        else:
            from sparcur import export as ex
            blob_ir = self._export(ex.ExportXml).latest_ir  # FIXME need to load?

        mbf_types = tuple(c.mimetype for c in
                          (exml.ExtractMBF, exml.ExtractNeurolucida))
        if self.options.unique:
            key = lambda p: (p[2], not p[0], p[1].lower(), p[1])
        else:
            key = lambda p: (p[2], p[3], not p[0], p[1].lower(), p[1])

        all_conts = sorted(set(((OntId(c['id_ontology'])
                                 if 'id_ontology' in c else
                                 ''),
                                c['name'],
                                settype(metadata_blob['type']),
                                *((dataset_xml['dataset_id'],)
                                  if not self.options.unique else et))
                               for dataset_xml in blob_ir.values()
                               if dataset_xml['type'] == 'all-xml-files'  # FIXME
                               for metadata_blob in dataset_xml['xml']
                               if metadata_blob['type'] in mbf_types and
                               'contours' in metadata_blob['extracted']
                               for c in metadata_blob['extracted']['contours']),
                           key=key)
        header = [['id', 'name', 'metadata source', 'dataset']]
        if self.options.unique:
            header[0] = header[0][:-1]

        return self._print_table(header + all_conts,
                                 title='Unique MBF contours',
                                 ext=ext)

    def terms(self, ext=None):
        # anatomy
        # cells
        # subcelluar
        import rdflib
        # FIXME cache these results and only recompute if latest changes?
        if self.options.raw:
            graph = self.summary.triples_exporter.graph
        else:
            from sparcur import export as ex  # FIXME very slow to import
            graph = OntGraph()
            self._export(ex.Export).latest_export_ttl_populate(graph)

        objects = set()
        skipped_prefixes = set()
        for t in graph:
            for e in t:
                if (isinstance(e, rdflib.URIRef) and
                    not e.startswith('info:') and
                    not e.startswith('doi:')):
                    oid = OntId(e)
                    if oid.prefix in want_prefixes:
                        objects.add(oid)
                    else:
                        skipped_prefixes.add(oid.prefix)

        if self.options.server and isinstance(ext, types.FunctionType):
            def reformat(ot):
                return [ot.label
                        if hasattr(ot, 'label') and ot.label else
                        '',
                        ot.atag(curie=True)]

        else:
            def reformat(ot):
                return [ot.label
                        if hasattr(ot, 'label') and ot.label else
                        '',
                        ot.curie]

        log.info(' '.join(sorted(skipped_prefixes)))
        objs = [OntTerm(o) if o.prefix not in ('TEMP', 'sparc') or
                o.prefix == 'TEMP' and o.suffix.isdigit() else
                o for o in objects]
        term_sets = {title:[o for o in objs if o.prefix == prefix]
                     for prefix, title in
                     (('NCBITaxon', 'Species'),
                      ('UBERON', 'Anatomy and age category'),  # FIXME
                      ('FMA', 'Anatomy (FMA)'),
                      ('PATO', 'Qualities'),
                      ('tech', 'Techniques'),
                      ('unit', 'Units'),
                      ('sparc', 'MIS terms'),
                      ('TEMP', 'Suggested terms'),
                     )}

        term_sets['Other'] = set(objs) - set(ot for v in term_sets.values()
                                             for ot in v)

        for title, terms in term_sets.items():
            header = [['Label', 'CURIE']]
            rows = header + [reformat(ot) for ot in
                            sorted(terms,
                                   key=lambda ot: (ot.prefix, ot.label.lower()
                                                   if hasattr(ot, 'label') and ot.label else ''))]

            yield self._print_table(rows, title=title, ext=ext)

    @idlib.utils.cache_result
    def _annos(self):
        group_id = auth.get('hypothesis-group')
        cache_file = Path(hyp.group_to_memfile(group_id + 'sparcur'))
        get_annos = hyp.Memoizer(memoization_file=cache_file, group=group_id)
        get_annos.api_token = auth.get('hypothesis-api-key')
        return get_annos()

    @idlib.utils.cache_result
    def _helpers(self):
        annos = self._annos()
        return [hyp.HypothesisHelper(a, annos) for a in annos]

    @idlib.utils.cache_result
    def _protc(self):
        from pysercomb.parsers import racket
        from protcur.analysis import protc
        annos = self._annos()
        return [protc(a, annos) for a in annos]

    def _protc_input(self, hrmd, tags, protc, ext):
        from pysercomb import exceptions as pexc
        from pysercomb.parsers import racket
        import pysercomb.pyr.units as pyru
        pyru.Hyp.bindImpl(None, HypothesisAnno=protc.byId)
        protcur_interpreter = pyru.Protc()
        def more(k, v):  # FIXME these can actually be parsed in parallel
            count = len(v)
            link = v[0].shareLink if count == 1 else ''
            facet = 'TODO'
            autos = set()
            for protc_helper in v:
                ok, expr, rest = racket.exp(repr(protc_helper))
                #log.critical(v)  # FIXME this induces an infinite loop I think due to a missing child
                #log.critical(expr)
                if expr is None or isinstance(expr, str):
                    continue
                try:
                    hrm = protcur_interpreter(expr)
                except pexc.ParseFailure as e:
                    log.error(e)
                    continue

                if isinstance(hrm.black_box, pyru.Term):
                    autos.add(hrm.black_box)

                continue

            auto_mapping = ';'.join([f'{bb.curie}|{bb.label}' for bb in autos])
            return count, link, facet, auto_mapping

        header = [['tag', 'value', 'text', 'exact', 'count', 'link', 'facet', 'auto-mapping']]
        #rows = [[(*k, ' '.join([a.shareLink for a in v]))] for k, v in sorted(hrm.items())]
        rows = [[*k, *more(k, v)]
                for k, v in sorted(list(hrmd.items()))]

        if self.options.sort_count_desc:
            rows = sorted(rows, key=lambda r: -r[-4])

        return self._print_table(header + rows,
                                 title=f'Annos for {" ".join(tags)}',
                                 ext=ext)

    def _process_protc(self, tags, matches, ext):
        from protcur.analysis import protc
        from pysercomb.pyr import units as pyru
        from pysercomb.parsers import racket

        lowernorm = ('protc:aspect',
                     'protc:implied-aspect',
                     'protc:executor-verb',
                     'protc:parameter*',
                     'protc:invariant',
                     )

        self._protc()
        # FIXME cannot breakpoint with this protc instances around ...
        pm = [protc.byId(a.id) for _, a in matches]

        # testing
        if 'protc:input' in tags:
            input = pm[13]
            invar = next(next(input.children).children)
            paramparser = pyru.ParamParser()
            tv = racket.sexp(invar.parameter())[1]
            tp = invar._parameter[1]
            assert tv == tp
            hrm = paramparser(tv)

        hrm = defaultdict(list)
        def denone(t): return tuple('' if e is None else e for e in t)
        for p in pm:
            pt = [t for t in p.tags if t in tags]
            if pt:
                hrm[denone((
                    pt[0],
                    normalize_exact(
                        p.value,
                        lower=anyMembers(tags, *lowernorm)),
                    normtext(
                        p.text,
                        lower=anyMembers(tags, *lowernorm)),
                    normalize_exact(
                        p.exact,
                        lower=anyMembers(tags,
                                            *lowernorm))))].append(p)
            else:
                log.warning(f'WAT {p.tags}')

        hrmd = dict(hrm)
        if 'protc:executor-verb' in tags:  # woo past tense
            ends_with_e = {h[1]:h for h, p in hrmd.items() if h[1].endswith('e')}
            others = {h[1]:h for h, p in hrmd.items()
                    if not h[1].endswith('ed') and not h[1].endswith('e')}
            for k, p in list(hrmd.items()):
                _, v, _, _ = k
                if v.endswith('ed'):
                    ewe = v[:-1]
                    if ewe in ends_with_e:
                        nk = ends_with_e[ewe]
                    elif v[:-2] in others:
                        nk = others[v[:-2]]
                    else:
                        continue

                    hrmd[nk].extend(hrmd.pop(k))

            # this has to come after due to remapping that happens above first
            wev = sheets.WorkingExecVerb()
            value_to_key, create = wev.condense()
            for _key in create:
                if _key not in hrmd:
                    hrmd[_key] = []

            for k, p in list(hrmd.items()):
                _, v, _, _ = k
                if v in value_to_key:
                    nk = value_to_key[v]
                    hrmd[nk].extend(hrmd.pop(k))

        if 'protc:input' in tags:
            return self._protc_input(hrmd, tags, protc, ext)

        def more(k, v):
            count = len(v)
            link = v[0].shareLink if count == 1 else ''
            facet = 'TODO'
            return count, link, facet

        header = [['tag', 'value', 'text', 'exact', 'count', 'link', 'facet']]
        #rows = [[(*k, ' '.join([a.shareLink for a in v]))] for k, v in sorted(hrm.items())]
        rows = [[*k, *more(k, v)]
                for k, v in sorted(hrmd.items())]

        if self.options.sort_count_desc:
            rows = sorted(rows, key=lambda r: -r[-3])

        return self._print_table(header + rows,
                                    title=f'Annos for {" ".join(tags)}',
                                    ext=ext)

    def _by_tags(self, *tags, links=True, ext=None):
        annos = self._annos()
        bannos = [ptcdoc.Annotation(a) for a in annos]  # better annos
        pool = ptcdoc.Pool(bannos)
        anno_counts = ptcdoc.AnnoCounts(pool)

        self._helpers()
        process_protc = anyMembers(tags,
                                   'protc:executor-verb',
                                   'protc:input',
                                   'protc:aspect',
                                   'protc:input-instance',
                                   'protc:implied-aspect',
                                   'protc:implied-input',
                                   'protc:parameter*',
                                   'protc:invariant',
                                   'protc:output',
                                   )

        # note that HypothesisHelper.byTags is an AND search not and OR search
        matches = [(normalize_exact(a.exact, lower=True), a)
                   for tag in tags for a in hyp.HypothesisHelper.byTags(tag)]

        if process_protc:
            return self._process_protc(tags, matches, ext)

        md = defaultdict(list)
        for e, a in matches:
            md[e].append(a)

        def gtag(e):  # FIXME memoize/cache?
            return ' '.join(sorted(set(t for a in md[e] for t in a.tags
                                       if t in tags)))

        def link(e): pass
        rows = sorted([[gtag(e),  # tag
                        e,        # key
                        c,        # count
                        (' '.join([a.htmlLink  # links
                                   if self.options.uri_api else
                                   a.shareLink for a in md[e]]
                                  if links else '')),
                        '',       # docs
                        '',       # exacts
                        ]
                       for e, c in Counter([normalize_exact(e)
                                            for e, a in matches]).most_common()],
                      key=lambda ab: (ab[0], -ab[2], ab[1].lower()))

        header = [['tag', 'key', 'count', 'links', 'docs', 'exacts']]
        return self._print_table(header + rows,
                                 title=f'Annos for {" ".join(tags)}',
                                 ext=ext)

    @sheets.AnnoTags.makeReportSheet()
    def anno_tags(self, *tags, links=False, ext=None):
        if not tags:
            tags = self.options.tag

        links = links or self.options.uri or self.options.uri_api

        return self._by_tags(*tags, links=links, ext=ext)

    @property
    @idlib.utils.cache_result
    def _ttlfile(self):
        # TODO uri/path reconciliation 
        if self.options.ttl_file:
            ontres = OntRes.fromStr(self.options.ttl_file)
        else:
            # FIXME make this go away and derive it from git
            iri = 'https://cassava.ucsd.edu/sparc/exports/curation-export.ttl'
            ontres = OntResIri(iri)

        if self.options.ttl_compare:
            ontres_compare_to = OntRes.fromStr(self.options.ttl_compare)
        else:
            ontres_compare_to = None

        return TtlFile(ontres, ontres_compare_to)

    def changes(self):
        tout = self._ttlfile.changes()

    @sheets.Reports.makeReportSheet('curies')
    def mis(self, ext=None):
        tout = self._ttlfile.mis(ext=ext)
        return self._print_table(tout,
                                 title=f'MIS predicates',
                                 ext=ext)
